(*----------------------------------------------------------------------*)
(* FILE: mk_S2L_L2Ssml                                      		*)
(* DESCRIPTION:                                                         *)
(*----------------------------------------------------------------------*) 

open Parse HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ; 


val _ = new_theory "S2L_L2S";

(*
load "bossLib";
load "some_list_thmsTheory";
load "setLib";
*)


open bossLib;
open setLib;
open some_list_thmsTheory;
open listTheory;
open rich_listTheory;
open setTheory;


val INDUCT_TAC = numLib.INDUCT_TAC;

fun DEL_TAC (t:term) = 
    UNDISCH_TAC t THEN DISCH_TAC
    THEN POP_ASSUM (fn thm =>  ALL_TAC) ;

val prove_thm = fn (name,tm,t) => save_thm(name, TAC_PROOF(([],tm),t));


val UNDISCH_ONE_TAC =
    let
        fun th_tac (th:thm) (tac:tactic) = (UNDISCH_TAC (concl th)) THEN tac
        fun u_asm  (th:thm) = itlist  th_tac [th] ALL_TAC
    in
        FIRST_ASSUM u_asm
    end;
 
val XRULE_ASSUM_TAC = 
       let 
        fun app (rule:thm->thm) = (fn (t:thm) => (rule t) handle _ =>  t)
       in
       RULE_ASSUM_TAC o app
       end;
  
fun REWRITE_ASM_TAC thms =
   XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE thms));
   
fun DO_TAC tac n = 
    if (n = 0)
    then ALL_TAC
    else tac THEN (DO_TAC tac (n-1));

fun POP_ASMS (n:int) =
  if (n=0)
  then ALL_TAC
  else POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASMS (n-1);

  
(*--------------------------------------------------------------------------*)  
(* Stuff about bijections from a subset of num to a set, for the            *)
(* characterisation of a function S2L, that converts a set to a list        *)
(*--------------------------------------------------------------------------*)  

val is_BIJ_NUM_to = new_definition ("is_BIJ_NUM_to",
   (--`is_BIJ_NUM_to (f:num->'a) (s:('a)set)
           = (!n m. ((n < (CARD s)) /\ (m < (CARD s))) ==> (f n = f m) ==> (n = m))
             /\
             (s = {f n | n < (CARD s)})`--));
                                   
val _ = set_fixity "is_BIJ_NUM_to" (Infixl 500);

val set_BIJ = new_definition ("set_BIJ",
  (--`set_BIJ (s:('a)set) = (@f.f is_BIJ_NUM_to s)`--));
                                   

    
val set_BIJ_DEF = store_thm("set_BIJ_DEF",
   (--`!(s:('a)set). (FINITE s) ==> ((set_BIJ s) is_BIJ_NUM_to s)`--),
   
   GEN_TAC THEN STRIP_TAC THEN REWRITE_TAC[set_BIJ]
   THEN IMP_RES_TAC (FINITE_ISO_NUM)
   THEN CONV_TAC SELECT_CONV
   THEN EXISTS_TAC (--`f:num->'a`--) THEN REWRITE_TAC[is_BIJ_NUM_to]
   THEN CONJ_TAC THEN FIRST_ASSUM MATCH_ACCEPT_TAC);


(* theorems about set_BIJ *)
   

val IN_s_IMP_EXISTS_n = prove(

    (--`!(x:'a) (s:('a)set).
            ((FINITE s) /\ (x IN s))  ==> ?n. (n < (CARD s) /\ (set_BIJ s n = x))`--),
            
     REPEAT STRIP_TAC  
     THEN IMP_RES_TAC (REWRITE_RULE[is_BIJ_NUM_to] set_BIJ_DEF)
     THEN UNDISCH_TAC (--`(x:'a) IN s`--) THEN ONCE_ASM_REWRITE_TAC[] 
     THEN CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV)
     THEN STRIP_TAC THEN EXISTS_TAC (--`n:num`--) THEN XRULE_ASSUM_TAC SYM
     THEN ASM_REWRITE_TAC[]);
     

val set_BIJ_s_IN_s = prove(
     (--`!(n:num) (s:('a)set). (FINITE s) /\ (n < (CARD s)) ==> ((set_BIJ s n) IN s)`--),
     
     REPEAT STRIP_TAC 
     THEN IMP_RES_TAC (REWRITE_RULE[is_BIJ_NUM_to] set_BIJ_DEF)
     THEN ONCE_ASM_REWRITE_TAC[] 
     THEN CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV)
     THEN EXISTS_TAC (--`n:num`--) THEN CONJ_TAC
     THENL
     [XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
     ,
      ONCE_ASM_REWRITE_TAC[]
     ]);



(*--------------------------------------------------------------------------*)   
(* interval m = [(m-1), ... , 2,1,0], needed for the conversion             *)
(* of sets to lists.                                                        *)     
(*--------------------------------------------------------------------------*)

    
val interval = new_recursive_definition
        {name = "interval",
         rec_axiom = prim_recTheory.num_Axiom,
         def = (--`(interval 0 = [])
                   /\
                   (interval (SUC n) = CONS n (interval n))`--)};


val LESS_n_EQ_IN_interval_n = store_thm ("LESS_n_EQ_IN_interval_n",
   (--`!(m:num) (n:num). (n < m) = (IS_EL n (interval m))`--),
   
  INDUCT_TAC
  THENL
   [STRIP_TAC THEN EQ_TAC
    THENL[REWRITE_TAC[prim_recTheory.NOT_LESS_0],REWRITE_TAC[interval,IS_EL]]
   ,
    STRIP_TAC THEN EQ_TAC
    THENL
    [REPEAT STRIP_TAC THEN REWRITE_TAC[interval, IS_EL]
     THEN REWRITE_ASM_TAC[prim_recTheory.LESS_THM]
     THEN UNDISCH_ONE_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
     THEN DISJ2_TAC THEN RES_TAC
    ,
     REWRITE_TAC[interval, IS_EL] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
     THENL[REWRITE_TAC[prim_recTheory.LESS_SUC_SUC],ALL_TAC]
     THEN MATCH_MP_TAC (arithmeticTheory.LESS_EQ_IMP_LESS_SUC)
     THEN MATCH_MP_TAC (arithmeticTheory.LESS_IMP_LESS_OR_EQ)
     THEN ASM_REWRITE_TAC[]
    ]
   ]);
   

val interval_LENGTH = store_thm("interval_LENGTH",
  (--`!(n:num). (LENGTH ((interval n):(num)list)) = n`--),
  
INDUCT_TAC THEN REWRITE_TAC[interval,LENGTH] THEN ASM_REWRITE_TAC[]);


val EL_interval = store_thm ("EL_interval",
   (--`!m n. (n < (SUC m)) ==> (EL n (interval (SUC m)) = m - n)`--),

INDUCT_TAC 
THENL
[REPEAT STRIP_TAC
THEN SUBGOAL_THEN (--`n=0`--) (fn thm => REWRITE_TAC[thm,interval,HD,EL,arithmeticTheory.SUB_0])
THEN UNDISCH_ONE_TAC THEN DECIDE_TAC
,
ALL_TAC
]
THEN INDUCT_TAC
THENL
[STRIP_TAC THEN REWRITE_TAC[interval,EL,HD,arithmeticTheory.SUB_0] 
,
 STRIP_TAC THEN ONCE_REWRITE_TAC[interval] THEN REWRITE_TAC[EL,TL] 
 THEN REWRITE_TAC[arithmeticTheory.SUB_MONO_EQ]
 THEN FIRST_ASSUM MATCH_MP_TAC THEN UNDISCH_ONE_TAC
 THEN DECIDE_TAC
]);




val LESS_0 = prim_recTheory.LESS_0;


val interval_REVERSE_from = store_thm("interval_REVERSE_from",
   (--`!n. (interval n) = (REVERSE (from 0 n))`--),
 
INDUCT_TAC 
THENL
[REWRITE_TAC[interval,from_def,REVERSE]
,
MATCH_MP_TAC LIST_EQ 
THEN REWRITE_TAC[from_LENGTH,interval_LENGTH,LENGTH_REVERSE]
THEN REPEAT STRIP_TAC 
THEN POP_ASSUM (fn thm => ASSUME_TAC thm THEN REWRITE_TAC [(REWRITE_RULE [from_LENGTH,thm] 
                               (ISPEC (--`from 0 (SUC n)`--) (ISPEC (--`n':num`--) EL_REVERSE)))])
THEN SUBGOAL_THEN (--`PRE (SUC n - n') < SUC n`--) ASSUME_TAC
THENL[REWRITE_TAC[arithmeticTheory.SUB]
THEN SUBGOAL_THEN (--`~(n < n')`--) (fn thm => REWRITE_TAC[thm])
THENL[UNDISCH_ONE_TAC THEN DECIDE_TAC,ALL_TAC]
THEN REWRITE_TAC[prim_recTheory.PRE] THEN DECIDE_TAC,ALL_TAC]
THEN POP_ASSUM (fn thm1 => ASSUME_TAC thm1 
                          THEN (POP_ASSUM (fn thm2 => 
                          REWRITE_TAC [(REWRITE_RULE [thm1,thm2,LESS_0,arithmeticTheory.ADD] 
                         (ISPEC (--`0`--) (ISPEC (--`(SUC n)`--) (ISPEC (--`PRE (SUC n - n')`--) EL_from))))])))
THEN REWRITE_TAC[arithmeticTheory.SUB]
THEN SUBGOAL_THEN (--`~(n < n')`--) (fn thm => REWRITE_TAC[thm])
THENL[UNDISCH_ONE_TAC THEN DECIDE_TAC,ALL_TAC]	THEN REWRITE_TAC[prim_recTheory.PRE]		       
THEN MATCH_MP_TAC EL_interval THEN ASM_REWRITE_TAC[]]);

 
 

val interval_split = store_thm("interval_split",
   (--`!k n. (k < n) ==>
       ((interval n) = (APPEND (APPEND (REVERSE (from (SUC k) (n - (SUC k)))) [k]) (interval k)))`--) ,
  
INDUCT_TAC THEN REWRITE_TAC[prim_recTheory.PRE,interval,APPEND_NIL]
THENL
[REPEAT STRIP_TAC
 THEN REWRITE_TAC[interval_REVERSE_from,GSYM SNOC_APPEND]
 THEN REWRITE_TAC[GSYM (CONJUNCT2 REVERSE)]
 THEN REWRITE_TAC[GSYM (CONJUNCT2 from_def)]
 THEN SUBGOAL_THEN (--`(SUC (n - (SUC 0))) = n`--) (fn thm => REWRITE_TAC[thm])
 THEN UNDISCH_ONE_TAC THEN DECIDE_TAC 
,
REPEAT STRIP_TAC 
THEN SUBGOAL_THEN (--`k < n`--) ASSUME_TAC
THENL[UNDISCH_ONE_TAC THEN DECIDE_TAC,ALL_TAC]
THEN RES_TAC THEN ASM_REWRITE_TAC[]
THEN CONV_TAC (RATOR_CONV (REWRITE_CONV[GSYM APPEND_ASSOC]))
THEN MATCH_MP_TAC APPEND_EQ THEN REWRITE_TAC[APPEND]
THEN REWRITE_TAC[GSYM SNOC_APPEND]

THEN SUBGOAL_THEN (--`(n - (SUC k)) = (SUC (n - (SUC (SUC k))))`--) ASSUME_TAC
THENL[UNDISCH_TAC (--`SUC k < n`--) THEN RW_TAC arith_ss [] ,ALL_TAC]
THEN ASM_REWRITE_TAC[from_def,REVERSE]  
]);
 


(*--------------------------------------------------------------------------*)
(* definition of S2L (set to list) and L2S (list to set)                    *)
(*--------------------------------------------------------------------------*)   

val S2L_DEF = new_definition ("S2L_DEF",
   (--`S2L (s:('a)set) = (MAP ((set_BIJ s):num -> 'a) (interval (CARD s)))`--));
   


val L2S_DEF = new_recursive_definition
        {name = "L2S_DEF",
         rec_axiom = list_Axiom,
         def = (--`(L2S ([]:('a)list) = ({}:('a)set))
                   /\
                   (L2S (CONS x xs) = (x INSERT (L2S xs)))`--)};


(*--------------------------------------------------------------------------*)
(* theorems about S2L and L2S                                               *)
(*--------------------------------------------------------------------------*)  
 
 
val S2L_EMPTY = store_thm("S2L_EMPTY",
  (--`S2L ({}:('a)set) = []`--),
  
  REWRITE_TAC[S2L_DEF,CARD_DEF,interval,MAP]);


val LIST_INDUCT_TAC = INDUCT_THEN list_INDUCT ASSUME_TAC;      

val IN_L2S_IS_EL = store_thm("IN_L2S_IS_EL",
  (--`!(x:'a) (xs:('a)list). (x IN (L2S xs)) = (IS_EL x xs)`--),
   STRIP_TAC THEN LIST_INDUCT_TAC
   THENL
    [REWRITE_TAC[L2S_DEF,NOT_IN_EMPTY,IS_EL]
    ,
     REWRITE_TAC[L2S_DEF, IS_EL,IN_INSERT]
     THEN ASM_REWRITE_TAC[]
    ]);
    

val IS_EL_MAP = prove(
  (--`!(xs:('a)list) (f:'a->'b) (y:'b).
          (?(x:'a). (IS_EL x xs) /\ ((f x) = y)) = (IS_EL y (MAP f xs))`--),

LIST_INDUCT_TAC THEN REWRITE_TAC[IS_EL]
THEN REWRITE_TAC [MAP,IS_EL]
THEN REPEAT STRIP_TAC THEN EQ_TAC THEN STRIP_TAC
THENL
 [DISJ1_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,
  DISJ2_TAC THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC GSYM THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`x':'a`--) THEN ASM_REWRITE_TAC[]
 ,
  EXISTS_TAC (--`x:'a`--) THEN ASM_REWRITE_TAC[]
 ,
  UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC GSYM THEN ASM_REWRITE_TAC[]
  THEN STRIP_TAC THEN  EXISTS_TAC (--`x':'a`--) THEN ASM_REWRITE_TAC[]
 ]);


val IN_IS_EL_S2L = store_thm("IN_IS_EL_S2L",

   (--`!(s:('a)set). (FINITE s) ==> 
          ((x IN s) = (IS_EL x (S2L s)))`--),
   
   REPEAT STRIP_TAC THEN IMP_RES_TAC set_BIJ_DEF
   THEN UNDISCH_ONE_TAC THEN ONCE_REWRITE_TAC[is_BIJ_NUM_to,S2L_DEF]
   THEN STRIP_TAC THEN EQ_TAC THEN STRIP_TAC
   THENL
    [REWRITE_TAC[GSYM IS_EL_MAP] 
     THEN SUBGOAL_THEN (--`(?n. (x = set_BIJ (s:('a)set) n) /\ n < CARD s)`--) ASSUME_TAC
     THENL[UNDISCH_TAC (--`(x:'a) IN s`--) THEN ONCE_ASM_REWRITE_TAC[] 
           THEN CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV) THEN STRIP_TAC
           THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`n:num`--)
           THEN ASM_REWRITE_TAC[], UNDISCH_ONE_TAC THEN STRIP_TAC]
     THEN EXISTS_TAC (--`n:num`--) 
     THEN CONJ_TAC THENL[ALL_TAC, CONV_TAC SYM_CONV THEN FIRST_ASSUM MATCH_ACCEPT_TAC]
     THEN REWRITE_TAC[GSYM LESS_n_EQ_IN_interval_n] THEN ASM_REWRITE_TAC[]
    ,
     ONCE_ASM_REWRITE_TAC[] THEN CONV_TAC (ONCE_DEPTH_CONV SET_SPEC_CONV)
     THEN REWRITE_TAC[LESS_n_EQ_IN_interval_n] THEN REWRITE_ASM_TAC[GSYM IS_EL_MAP]
     THEN UNDISCH_ONE_TAC THEN STRIP_TAC THEN EXISTS_TAC (--`x':num`--) THEN ASM_REWRITE_TAC[]
    ]);
    

val L2S_MAP_eq_IMAGE = store_thm("L2S_MAP_eq_IMAGE",
  (--`!(tl:('a)list) (f:'a->'b).
     ((L2S (MAP f tl)) = (IMAGE f (L2S tl)))`--),
     
  LIST_INDUCT_TAC
  THENL
   [REWRITE_TAC[MAP,L2S_DEF,IMAGE_EMPTY]
   ,
    ASM_REWRITE_TAC[MAP,L2S_DEF,IMAGE_INSERT]
    ]);


      
    


val EXISTS_LIST_FOR_S2L = store_thm("EXISTS_LIST_FOR_S2L",
  (--`!(s:('a)set) . ?(tl:('a)list). tl = (S2L s)`--),
  
  STRIP_TAC THEN EXISTS_TAC (--`S2L (s:('a)set)`--)
  THEN REFL_TAC);
  
val L2S_S2L_id = store_thm("L2S_S2L_id",
   (--`!(s:('a)set). (FINITE s) ==> ((L2S (S2L s)) = s)`--),
   
    SET_INDUCT_TAC
    THENL
     [REWRITE_TAC[S2L_EMPTY,L2S_DEF]
     ,
      REWRITE_TAC[S2L_DEF] THEN IMP_RES_TAC CARD_INSERT
      THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[interval,MAP]
      THEN REWRITE_TAC[EXTENSION,IN_INSERT,L2S_DEF]
      THEN STRIP_TAC THEN EQ_TAC 
      THENL
       [STRIP_TAC THEN ASM_CASES_TAC (--`(x:'a) = e`--)
        THENL[ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[], DISJ2_TAC,
              ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[],DISJ2_TAC]
        THENL
         [ASM_REWRITE_TAC[] 
          THEN SUBGOAL_THEN (--`FINITE (e INSERT (s:('a)set))`--) ASSUME_TAC
          THENL
           [ASM_REWRITE_TAC[FINITE_INSERT]
           ,DEL_TAC (--`FINITE (s:('a)set)`--)]
          THEN SUBGOAL_THEN (--`(CARD (s:('a)set)) < (CARD (e INSERT s))`--) ASSUME_TAC
          THENL[ASM_REWRITE_TAC[prim_recTheory.LESS_SUC_SUC],ALL_TAC]
          THEN IMP_RES_TAC (ISPEC (--`(e:'a) INSERT s`--) (ISPEC (--`(CARD (s:('a)set))`--) set_BIJ_s_IN_s))
          THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[IN_INSERT] THEN XRULE_ASSUM_TAC SYM
          THEN ONCE_ASM_REWRITE_TAC[] THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[]
         ,
          XRULE_ASSUM_TAC (REWRITE_RULE [IN_L2S_IS_EL])
          THEN XRULE_ASSUM_TAC (REWRITE_RULE [GSYM IS_EL_MAP])
          THEN  UNDISCH_ONE_TAC  THEN  UNDISCH_ONE_TAC 
	  THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) 
          THEN REPEAT STRIP_TAC
          THEN ASM_REWRITE_TAC[]
          THEN SUBGOAL_THEN (--`FINITE (e INSERT (s:('a)set))`--) ASSUME_TAC
          THENL
           [ASM_REWRITE_TAC[FINITE_INSERT]
           ,DEL_TAC (--`FINITE (s:('a)set)`--)]
          THEN SUBGOAL_THEN (--`x' < (CARD ((e:'a) INSERT s))`--) ASSUME_TAC
          THENL[XRULE_ASSUM_TAC (REWRITE_RULE [GSYM LESS_n_EQ_IN_interval_n])
                THEN ASM_REWRITE_TAC[] 
                THEN MATCH_MP_TAC (prim_recTheory.LESS_SUC)
                THEN ASM_REWRITE_TAC[],ALL_TAC] 
          THEN IMP_RES_TAC (ISPEC (--`(e:'a) INSERT s`--) (ISPEC (--`x':num`--) set_BIJ_s_IN_s))
          THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[IN_INSERT] 
          THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`~((e:'a) = x)`--)
          THEN ASM_REWRITE_TAC[]
         ]
       ,
        STRIP_TAC 
        THEN ASM_CASES_TAC (--`(set_BIJ (e INSERT s) (CARD s)=(x:('a)))`--)
        THENL[ASM_REWRITE_TAC[],DISJ2_TAC,ASM_REWRITE_TAC[],DISJ2_TAC]
        THENL
         [REWRITE_TAC[IN_L2S_IS_EL] THEN REWRITE_TAC [GSYM IS_EL_MAP]
         THEN REWRITE_TAC [GSYM LESS_n_EQ_IN_interval_n]
         THEN SUBGOAL_THEN (--`FINITE (e INSERT (s:('a)set))`--) ASSUME_TAC
          THENL
           [ASM_REWRITE_TAC[(FINITE_INSERT)]
           ,DEL_TAC (--`FINITE (s:('a)set)`--)]
          THEN SUBGOAL_THEN (--`e IN ((e:'a) INSERT s)`--) ASSUME_TAC
          THENL[REWRITE_TAC[IN_INSERT],ALL_TAC]
          THEN IMP_RES_TAC IN_s_IMP_EXISTS_n THEN EXISTS_TAC (--`n':num`--)
          THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`n' < CARD ((e:'a) INSERT s)`--)
          THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[prim_recTheory.LESS_THM] 
          THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV)
          THEN REPEAT STRIP_TAC
          THEN UNDISCH_TAC (--`~(set_BIJ ((e:'a) INSERT s) (CARD s) = x)`--) THEN ASM_REWRITE_TAC[]
         ,
          REWRITE_TAC[IN_L2S_IS_EL] THEN REWRITE_TAC [GSYM IS_EL_MAP]
         THEN REWRITE_TAC [GSYM LESS_n_EQ_IN_interval_n]
         THEN SUBGOAL_THEN (--`FINITE (e INSERT (s:('a)set))`--) ASSUME_TAC
          THENL
           [ASM_REWRITE_TAC[(FINITE_INSERT)]
           ,DEL_TAC (--`FINITE (s:('a)set)`--)]
          THEN SUBGOAL_THEN (--`x IN ((e:'a) INSERT s)`--) ASSUME_TAC
          THENL[ASM_REWRITE_TAC[IN_INSERT],ALL_TAC]
          THEN IMP_RES_TAC IN_s_IMP_EXISTS_n THEN EXISTS_TAC (--`n':num`--)
          THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`n' < CARD ((e:'a) INSERT s)`--)
          THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[prim_recTheory.LESS_THM] 
          THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV)
          THEN REPEAT STRIP_TAC
          THEN UNDISCH_TAC (--`~(set_BIJ ((e:'a) INSERT s) (CARD s) = x)`--) THEN ASM_REWRITE_TAC[]
         ]
       ]
     ]);
     


val FINITE_IMP_FINITE_L2S_S2L = store_thm("FINITE_IMP_FINITE_L2S_S2L",

  (--`!(s:('a)set). (FINITE s) ==> (FINITE (L2S (S2L s)))`--),
  
  REPEAT STRIP_TAC THEN IMP_RES_TAC L2S_S2L_id THEN ASM_REWRITE_TAC[]);
  

val L2S_FINITE = store_thm("L2S_FINITE",
(--`!(xs:('a)list). (FINITE (L2S xs))`--),
     
 REWRITE_TAC[FINITE_DEF] THEN LIST_INDUCT_TAC
 THEN REWRITE_TAC[L2S_DEF]
 THENL
  [REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
  ,
   REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC 
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[]
  ]  );
  
	   

val NO_DUPLICATES_S2L = store_thm("NO_DUPLICATES_S2L",
   (--`!(s:('a) set). (FINITE s) ==> NO_DUPLICATES (S2L s)`--),

SET_INDUCT_TAC	  
THENL[REWRITE_TAC[S2L_EMPTY,NO_DUPLICATES_def,LENGTH,prim_recTheory.NOT_LESS_0],ALL_TAC]
THEN REWRITE_TAC[S2L_DEF,NO_DUPLICATES_def] THEN REPEAT GEN_TAC THEN STRIP_TAC
THEN REWRITE_ASM_TAC[LENGTH_MAP] 
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==}, 
                             {redex = ==`:'b`==, residue = ==`:'a`==}] EL_MAP)
THEN ASM_REWRITE_TAC[]
THEN SUBGOAL_THEN (--`(CARD ((e:'a) INSERT s)) = (SUC (CARD s))`--) ASSUME_TAC
THENL
 [IMP_RES_TAC CARD_INSERT THEN ASM_REWRITE_TAC[],
  POP_ASSUM (fn thm => REWRITE_TAC[thm] THEN REWRITE_ASM_TAC[thm] THEN ASSUME_TAC thm)]
THEN ASSUME_TAC (SYM (ISPEC (--`s:('a)set`--) (ISPEC (--`e:'a`--) (FINITE_INSERT))))
THEN UNDISCH_TAC (--`FINITE (s:('a)set)`--) THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
THEN IMP_RES_TAC set_BIJ_DEF THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[is_BIJ_NUM_to]
THEN REWRITE_ASM_TAC [interval_LENGTH] THEN ASM_REWRITE_TAC[]
THEN IMP_RES_TAC EL_interval THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
THEN XRULE_ASSUM_TAC (ISPEC (--`(CARD (s:('a)set)) - n`--))
THEN XRULE_ASSUM_TAC (ISPEC (--`(CARD (s:('a)set)) - k`--))
THEN STRIP_TAC
THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
THEN UNDISCH_TAC (--`~((n:num) = k)`--)
THEN SUBGOAL_THEN (--`(((CARD (s:('a)set)) - n) = ((CARD s) - k)) = (n = k)`--)
      (fn thm => REWRITE_TAC[SYM thm])
THENL[MATCH_MP_TAC (arithmeticTheory.SUB_CANCEL)
THEN UNDISCH_TAC (--`n < SUC (CARD (s:('a)set))`--) THEN UNDISCH_TAC (--`k < SUC (CARD (s:('a)set))`--)
THEN RW_TAC arith_ss [],ALL_TAC]
THEN FIRST_ASSUM MATCH_MP_TAC 
THEN UNDISCH_TAC (--`n < SUC (CARD (s:('a)set))`--) THEN UNDISCH_TAC (--`k < SUC (CARD (s:('a)set))`--)
THEN RW_TAC arith_ss []);




val IS_EL_DEL_S2L = store_thm("IS_EL_DEL_S2L",
  (--`!(s:('a)set) (x:'a) (y:'a).
    ((FINITE s) /\ (IS_EL x (S2L s)) /\ (IS_EL y (DEL x (S2L s)))) ==> (~(x = y))`--),
    
  REPEAT STRIP_TAC THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
  THEN IMP_RES_TAC NO_DUPLICATES_S2L THEN IMP_RES_TAC NO_DUPLICATES_THM
  THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]);
   
val IS_EL_DEL_S2L_THM = store_thm("IS_EL_DEL_S2L_THM",
  (--`!(s:('a)set) (x:'a) (y:'a).
    ((FINITE s) /\ (IS_EL x (S2L s)) /\ (IS_EL y (DEL x (S2L s)))) ==> ((y IN s) /\ (~(x = y)))`--),
    
  REPEAT STRIP_TAC THENL[ALL_TAC, POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
  THEN IMP_RES_TAC NO_DUPLICATES_S2L THEN IMP_RES_TAC NO_DUPLICATES_THM
  THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]]
  THEN UNDISCH_ONE_TAC THEN IMP_RES_TAC (ISPEC (--`(S2L (s:('a)set))`--) split_LIST) 
  THEN ASM_REWRITE_TAC[IS_EL_APPEND] THEN STRIP_TAC
  THENL 
    let 
      val tac = 
        SUBGOAL_THEN (--`((y:'a) IN s) = (IS_EL y (S2L s)) `--) (fn thm => REWRITE_TAC[thm])
        THENL[MATCH_MP_TAC (GEN_ALL IN_IS_EL_S2L) THEN ASM_REWRITE_TAC[],ALL_TAC]
        THEN ASM_REWRITE_TAC[IS_EL_APPEND]
    in [tac,tac] end  
   );



val S2L_split = store_thm("S2L_split",

(--`(!(s:('a)set). (FINITE s) ==>
    (!x. (~(x IN s)) ==>
         (?(xs:('a)list) (ys:('a)list).
	  ((S2L (x INSERT s)) = (APPEND (APPEND xs [x]) ys))
           /\
	   (s = (L2S (APPEND xs ys))))))`--),
          
SET_INDUCT_TAC
THENL
[REWRITE_TAC[NOT_IN_EMPTY,S2L_EMPTY,S2L_DEF, MAP,
             CARD_DEF,CARD_SING, interval, 
	     prove((--`1 = (SUC 0)`--), DECIDE_TAC)]
THEN STRIP_TAC THEN EXISTS_TAC (--`[]:('a)list`--) 
THEN EXISTS_TAC (--`[]:('a)list`--)  THEN REWRITE_TAC[APPEND,L2S_DEF]

THEN ASSUME_TAC (ISPEC (--`x:'a`--) SING)
THEN IMP_RES_TAC (SING_FINITE) 
THEN IMP_RES_TAC IN_s_IMP_EXISTS_n THEN XRULE_ASSUM_TAC (ISPEC (--`x:'a`--))
THEN UNDISCH_ONE_TAC 
THEN REWRITE_TAC[IN_SING, CARD_SING, 
                      prove((--`1 = (SUC 0)`--), DECIDE_TAC),
                      prove((--`!n. (n < (SUC 0)) = (n = 0)`--),DECIDE_TAC) ] 
THEN STRIP_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
THEN ASM_REWRITE_TAC[]
,
REWRITE_TAC[IN_INSERT,DE_MORGAN_THM] THEN REPEAT STRIP_TAC      
THEN ASSUME_TAC (GSYM(ISPEC (--`s:('a)set`--) (ISPEC (--`e:'a`--) (FINITE_INSERT))))
THEN UNDISCH_TAC (--`FINITE (s:('a)set)`--) THEN ASM_REWRITE_TAC[] THEN STRIP_TAC    
THEN ASSUME_TAC (GSYM(ISPEC (--`(e INSERT (s:('a)set))`--) (ISPEC (--`x:'a`--) (FINITE_INSERT))))
THEN UNDISCH_TAC (--`FINITE (e INSERT (s:('a)set))`--) THEN ASM_REWRITE_TAC[] THEN STRIP_TAC    
THEN ASSUME_TAC (ISPEC (--`x:'a`--) (ISPEC (--`(e INSERT (s:('a)set))`--) (GEN_ALL (COMPONENT))))
THEN IMP_RES_TAC IN_s_IMP_EXISTS_n THEN POP_ASSUM (fn thm => ALL_TAC)
THEN POP_ASSUM (fn thm => ALL_TAC) 
THEN EXISTS_TAC (--`MAP (set_BIJ ((x:'a) INSERT e INSERT s)) 
                    (REVERSE(from (SUC n) ((CARD ((x:'a) INSERT e INSERT s)) - (SUC n))))`--)
THEN EXISTS_TAC (--`MAP (set_BIJ ((x:'a) INSERT e INSERT s)) (interval n)`--)
THEN CONJ_TAC
THENL
 [SUBGOAL_THEN (--`[(x:'a)] = [set_BIJ (x INSERT e INSERT s) (n:num)]`--) (fn thm => REWRITE_TAC[thm])
  THENL[ASM_REWRITE_TAC[],ALL_TAC]
  THEN REWRITE_TAC[GSYM SNOC_APPEND,GSYM MAP_SNOC,GSYM MAP_APPEND]
  THEN REWRITE_TAC[SNOC_APPEND] THEN IMP_RES_TAC interval_split
  THEN ASM_REWRITE_TAC[S2L_DEF]
 ,
 REWRITE_TAC[GSYM MAP_APPEND,L2S_MAP_eq_IMAGE,EXTENSION,IN_IMAGE]
 THEN REPEAT STRIP_TAC THEN EQ_TAC THEN STRIP_TAC
 THENL
  [ASM_CASES_TAC (--`(x':'a) = e`--)
   THENL
    [POP_ASSUM (fn thm => REWRITE_TAC[thm] THEN REWRITE_ASM_TAC[thm])
    THEN ASSUME_TAC 
     (ONCE_REWRITE_RULE[INSERT_COMM]
        (ISPEC (--`e:'a`--) (ISPEC (--`(x INSERT (s:('a)set))`--) (GEN_ALL (COMPONENT)))))
    THEN IMP_RES_TAC IN_s_IMP_EXISTS_n THEN DO_TAC (POP_ASSUM (fn thm => ALL_TAC)) 6
    THEN EXISTS_TAC (--`n':num`--) THEN ASM_REWRITE_TAC[IN_L2S_IS_EL,IS_EL_APPEND,IS_EL_REVERSE]
    THEN ASM_CASES_TAC (--`n' < n`--)
    THENL
     [DISJ2_TAC THEN ASM_REWRITE_TAC[GSYM LESS_n_EQ_IN_interval_n]
     ,
      DISJ1_TAC 
      THEN SUBGOAL_THEN (--`~(n = (n':num))`--) ASSUME_TAC
      THENL[STRIP_TAC THEN UNDISCH_TAC (--`set_BIJ ((x:'a) INSERT e INSERT s) n = x`--)
            THEN ASM_REWRITE_TAC[] THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV)
	    THEN ASM_REWRITE_TAC[],ALL_TAC]
      THEN MATCH_MP_TAC IS_EL_from2 THEN DO_TAC UNDISCH_ONE_TAC 2
      THEN UNDISCH_TAC (--`n' < CARD ((x:'a) INSERT e INSERT s)`--)
      THEN UNDISCH_TAC (--`n < CARD ((x:'a) INSERT e INSERT s)`--)
      THEN RW_TAC arith_ss []
     ]
   ,
    SUBGOAL_THEN (--`(x':'a) IN (x INSERT e INSERT s)`--) 
         (fn thm => ASSUME_TAC (REWRITE_RULE [thm] (ISPEC (--`((x:'a) INSERT e INSERT s)`--)
	                       (ISPEC (--`x':'a`--) IN_s_IMP_EXISTS_n))))
    THENL[ASM_REWRITE_TAC[IN_INSERT],ALL_TAC]
    THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
    THEN EXISTS_TAC (--`n':num`--) THEN ASM_REWRITE_TAC[IN_L2S_IS_EL,IS_EL_APPEND,IS_EL_REVERSE]
    THEN ASM_CASES_TAC (--`n' < n`--)
    THENL
     [DISJ2_TAC THEN ASM_REWRITE_TAC[GSYM LESS_n_EQ_IN_interval_n]
     ,
      DISJ1_TAC 
      THEN SUBGOAL_THEN (--`~(n = (n':num))`--) ASSUME_TAC
      THENL[SUBGOAL_THEN (--`~((x':'a) = x)`--) ASSUME_TAC
            THENL[STRIP_TAC THEN UNDISCH_TAC (--`~((x:'a) IN s)`--)
	    THEN UNDISCH_TAC (--`(x':'a) IN e INSERT s`--) THEN ASM_REWRITE_TAC[IN_INSERT]
	    ,ALL_TAC]
	    THEN STRIP_TAC THEN UNDISCH_TAC (--`~((x':'a) = x)`--)
	    THEN XRULE_ASSUM_TAC SYM THEN REWRITE_TAC[] THEN ONCE_ASM_REWRITE_TAC[]
	    THEN POP_ASSUM (fn thm => REWRITE_TAC[thm]),ALL_TAC]
      THEN MATCH_MP_TAC IS_EL_from2 THEN DO_TAC UNDISCH_ONE_TAC 2
      THEN UNDISCH_TAC (--`n' < CARD ((x:'a) INSERT e INSERT s)`--)
      THEN UNDISCH_TAC (--`n < CARD ((x:'a) INSERT e INSERT s)`--)
      THEN RW_TAC arith_ss []
     ]
    
   ]
   ,
       
    REWRITE_ASM_TAC[IN_L2S_IS_EL,IS_EL_APPEND,IS_EL_REVERSE]
    THEN REWRITE_ASM_TAC[GSYM LESS_n_EQ_IN_interval_n,IS_EL_from_EQ]
    THEN SUBGOAL_THEN (--`~((x'':num) = n)`--) ASSUME_TAC
    THENL[UNDISCH_ONE_TAC THEN RW_TAC arith_ss [],ALL_TAC]
    THEN SUBGOAL_THEN (--`~((x':'a) = x)`--) ASSUME_TAC
    THENL
     [STRIP_TAC THEN UNDISCH_TAC (--`set_BIJ ((x:'a) INSERT e INSERT s) n = x`--)
      THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm]) THEN REWRITE_TAC[]
      THEN UNDISCH_TAC (--`(x:'a) = set_BIJ (x INSERT e INSERT s) x''`--) THEN STRIP_TAC
      THEN POP_ASSUM (fn thm => CONV_TAC (RAND_CONV (RAND_CONV (ONCE_REWRITE_CONV[thm]))))
      THEN IMP_RES_TAC set_BIJ_DEF THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[is_BIJ_NUM_to,AND_IMP_INTRO]
      THEN REPEAT STRIP_TAC THEN UNDISCH_TAC (--`~(x'' = (n:num))`--) THEN REWRITE_TAC[]
      THEN CONV_TAC SYM_CONV THEN FIRST_ASSUM MATCH_MP_TAC 
      THEN POP_ASSUM (fn thm => REWRITE_TAC[thm]) THEN POP_ASSUM (fn thm => ALL_TAC)
      THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC
      THEN UNDISCH_ONE_TAC THEN DECIDE_TAC,ALL_TAC]
    THEN SUBGOAL_THEN (--`x' IN ((x:'a) INSERT e INSERT s)`--) ASSUME_TAC
    THENL[ASM_REWRITE_TAC[] THEN MATCH_MP_TAC set_BIJ_s_IN_s THEN ASM_REWRITE_TAC[] 
    THEN POP_ASSUM (fn thm => ALL_TAC) 
    THEN UNDISCH_TAC (--`n < CARD ((x:'a) INSERT e INSERT s)`--)
    THEN DO_TAC UNDISCH_ONE_TAC 2 THEN DECIDE_TAC,ALL_TAC]
    THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[IN_INSERT]
    
    
  ]
 ]
]
);

val CARD_EQ_LENGTH_S2L = prove(
  (--`!(s:('a)set). (FINITE s) ==> (\s. ((CARD s) = (LENGTH (S2L s)))) s`--),
     
MATCH_MP_TAC (FINITE_INDUCT)  
THEN BETA_TAC THEN REWRITE_TAC[S2L_EMPTY, LENGTH, CARD_EMPTY]
THEN REPEAT STRIP_TAC THEN STRIP_ASSUME_TAC (CARD_DEF)
THEN RES_TAC THEN ASM_REWRITE_TAC[] THEN XRULE_ASSUM_TAC SYM 
THEN ASM_REWRITE_TAC[S2L_DEF] 
THEN REWRITE_TAC[LENGTH_MAP, interval_LENGTH ]);


val CARD_EQ_LENGTH_S2L = prove_thm ("CARD_EQ_LENGTH_S2L",
  (--`!(s:('a)set). (FINITE s) ==> ((CARD s) = (LENGTH (S2L s)))`--),

MATCH_ACCEPT_TAC (BETA_RULE CARD_EQ_LENGTH_S2L));

   
     
 
val L2S_DEL = prove_thm("L2S_DEL",
 (--`!(s:('a)set). (FINITE s) ==> 
  (!(x:'a). (L2S (DEL x (S2L s))) = (s DELETE x))`--),

SET_INDUCT_TAC THEN REWRITE_TAC[L2S_DEF,S2L_EMPTY,DEL_def,EMPTY_DELETE]  
THEN STRIP_TAC THEN REWRITE_TAC[EXTENSION]  
THEN GEN_TAC THEN REWRITE_TAC[IN_L2S_IS_EL, IN_DELETE, IN_INSERT]
THEN SUBGOAL_THEN (--`FINITE ((e:'a) INSERT s)`--) ASSUME_TAC
THENL[REWRITE_TAC[(FINITE_INSERT)],ALL_TAC]
THEN IMP_RES_TAC NO_DUPLICATES_S2L THEN IMP_RES_TAC IS_EL_DEL THEN ASM_REWRITE_TAC[]
THEN IMP_RES_TAC (GSYM IN_IS_EL_S2L) THEN ASM_REWRITE_TAC[IN_INSERT]
THEN EQ_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]);


 
 
 

val FOLDR_MAP_EQ_GEN = prove_thm("FOLDR_MAP_EQ_GEN",
  (--`!(xs:('a)list) (ys:('a)list) (op:'b->'b->'b) (e:'b) (f:'a->'b).
     ((MONOID op e) /\ (COMM op) /\ (NO_DUPLICATES xs) /\ (NO_DUPLICATES ys)
      /\ ((L2S xs) = (L2S ys)) /\ (!(x:'a). (IS_EL x xs) ==> (f x = g x)))
     ==>
     (FOLDR op e (MAP f xs) = FOLDR op e (MAP g ys))`--),
     

LIST_INDUCT_TAC 
THENL
 [LIST_INDUCT_TAC THEN REWRITE_TAC[L2S_DEF, NOT_EMPTY_INSERT, MAP], ALL_TAC]
THEN GEN_TAC 
THEN REWRITE_TAC[FOLDR, MAP] THEN REWRITE_TAC[EXTENSION,IN_L2S_IS_EL]
THEN REPEAT STRIP_TAC
THEN SUBGOAL_THEN (--`IS_EL (x:'a) ys`--) ASSUME_TAC
THENL
 [POP_ASSUM (fn thm => (POP_ASSUM (fn thm2 =>  REWRITE_TAC[GSYM thm2] THEN ASSUME_TAC thm2)) THEN ASSUME_TAC thm)
  THEN REWRITE_TAC[IS_EL]
 ,ALL_TAC]
THEN IMP_RES_TAC split_LIST THEN ASM_REWRITE_TAC[MAP_APPEND]
THEN IMP_RES_TAC FOLDR_COMM_MONOID_APPEND THEN ASM_REWRITE_TAC[]
THEN REWRITE_TAC[FOLDR, MAP] THEN REWRITE_ASM_TAC[operatorTheory.MONOID_DEF,operatorTheory.RIGHT_ID_DEF]
THEN ASM_REWRITE_TAC[]
THEN SUBGOAL_THEN (--`COMM (op:'b->'b->'b)`--) ASSUME_TAC
THENL[ASM_REWRITE_TAC[],ALL_TAC]
THEN POP_ASSUM (fn thm => 
           CONV_TAC (RAND_CONV (RATOR_CONV (ONCE_REWRITE_CONV[REWRITE_RULE[operatorTheory.COMM_DEF] thm]))))
THEN SUBGOAL_THEN (--`ASSOC (op:'b->'b->'b)`--) ASSUME_TAC
THENL[ASM_REWRITE_TAC[],ALL_TAC]
THEN POP_ASSUM (fn thm => CONV_TAC (ONCE_REWRITE_CONV[GSYM (REWRITE_RULE[operatorTheory.ASSOC_DEF] thm)]))
THEN SUBGOAL_THEN (--`((f:'a->'b) (x:'a)) = (g x)`--) (fn thm => REWRITE_TAC[thm])
THENL[FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[IS_EL]
,ALL_TAC]
THEN AP_TERM_TAC 
THEN POP_ASSUM (fn thm => REWRITE_TAC[GSYM thm]) 
THEN REWRITE_TAC[GSYM (MAP_APPEND)]
THEN FIRST_ASSUM MATCH_MP_TAC  THEN ASM_REWRITE_TAC[]
THEN IMP_RES_TAC NO_DUPLICATES_CONS THEN ASM_REWRITE_TAC[]
THEN SUBGOAL_THEN (--`NO_DUPLICATES (APPEND (ys':('a)list) zs)`--) ASSUME_TAC
THENL
 [UNDISCH_TAC (--`NO_DUPLICATES (ys:('a)list)`--) THEN ASM_REWRITE_TAC[NO_DUPLICATES_def]
  THEN REWRITE_TAC[LENGTH_APPEND, LENGTH] THEN STRIP_TAC 
  THEN REPEAT GEN_TAC THEN STRIP_TAC
  THEN STRIP_ASSUME_TAC (ISPEC (--`LENGTH (ys':('a)list)`--) (ISPEC (--`n:num`--) (arithmeticTheory.LESS_CASES)))
  THENL
   [IMP_RES_TAC (EL_APPEND1) THEN ASM_REWRITE_TAC[]
    THEN STRIP_ASSUME_TAC (ISPEC (--`LENGTH (ys':('a)list)`--) (ISPEC (--`k:num`--) (arithmeticTheory.LESS_CASES)))
    THENL
     [IMP_RES_TAC (EL_APPEND1) THEN ASM_REWRITE_TAC[]
      THEN XRULE_ASSUM_TAC (ISPEC (--`n:num`--)) THEN XRULE_ASSUM_TAC (ISPEC (--`k:num`--)) 
      THEN UNDISCH_TAC (--`n < (LENGTH ys' + SUC 0) + LENGTH zs /\ k < (LENGTH (ys':('a)list) + SUC 0) + LENGTH zs /\
                                    ~(n = k) ==> ~(EL n (APPEND (APPEND ys' [x]) zs) = EL k (APPEND (APPEND
				    ys' [x]) zs))`--)
      THEN ASM_REWRITE_TAC[GSYM(APPEND_ASSOC)] THEN STRIP_TAC
      THEN FIRST_ASSUM MATCH_MP_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
      THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN DECIDE_TAC
    ,
      IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
      THEN XRULE_ASSUM_TAC (ISPEC (--`n:num`--)) THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC k):num`--)) 
      THEN UNDISCH_TAC (--`n < (LENGTH ys' + SUC 0) + LENGTH zs /\ (SUC k) < (LENGTH (ys':('a)list) + SUC 0) + LENGTH zs /\
                                    ~(n = (SUC k)) ==> ~(EL n (APPEND (APPEND ys' [x]) zs) = EL (SUC k)
				    (APPEND (APPEND ys' [x]) zs))`--)
      THEN ASM_REWRITE_TAC[GSYM(APPEND_ASSOC)] 
      THEN SUBGOAL_THEN (--`LENGTH (ys':('a)list) <= (SUC k)`--) ASSUME_TAC
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= k`--) THEN DECIDE_TAC,ALL_TAC]
      THEN IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
      THEN SUBGOAL_THEN (--`(SUC k) - (LENGTH (ys':('a)list)) = (SUC (k - (LENGTH (ys':('a)list))))`--) (fn thm => REWRITE_TAC[thm])
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= k`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN REWRITE_TAC[EL, APPEND, TL]
      THEN STRIP_TAC 
      THEN FIRST_ASSUM MATCH_MP_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
      THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC 
      THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC 
      THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN RW_TAC arith_ss []
    ]
   ,
    IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
    THEN STRIP_ASSUME_TAC (ISPEC (--`LENGTH (ys':('a)list)`--) (ISPEC (--`k:num`--) (arithmeticTheory.LESS_CASES)))
    THENL
     [IMP_RES_TAC (EL_APPEND1) THEN ASM_REWRITE_TAC[]
      THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC n):num`--)) THEN XRULE_ASSUM_TAC (ISPEC (--`k:num`--)) 
      THEN UNDISCH_TAC (--`SUC n < (LENGTH (ys':('a)list) + SUC 0) + LENGTH zs /\ k < (LENGTH ys' + SUC 0) + LENGTH zs /\
          ~(SUC n = k) ==>  ~(EL (SUC n) (APPEND (APPEND ys' [x]) zs) 
	  = EL k (APPEND (APPEND ys' [x]) zs))`--)
      
      THEN ASM_REWRITE_TAC[GSYM(APPEND_ASSOC)]       
      THEN SUBGOAL_THEN (--`LENGTH (ys':('a)list) <= (SUC n)`--) ASSUME_TAC
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= n`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
      THEN SUBGOAL_THEN (--`(SUC n) - (LENGTH (ys':('a)list)) = (SUC (n - (LENGTH (ys':('a)list))))`--) (fn thm => REWRITE_TAC[thm])
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= n`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN REWRITE_TAC[EL, APPEND, TL]
      THEN STRIP_TAC 
      THEN FIRST_ASSUM MATCH_MP_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
      THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC 
      THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC 
      THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN RW_TAC arith_ss []
     ,
      IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
      THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC n):num`--)) THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC k):num`--)) 
      THEN UNDISCH_TAC (--`SUC n < (LENGTH (ys':('a)list) + SUC 0) + LENGTH zs /\ (SUC k) < (LENGTH ys' + SUC 0) + LENGTH zs /\
          ~(SUC n = (SUC k)) ==>  ~(EL (SUC n) (APPEND (APPEND ys' [x]) zs) 
	  = EL (SUC k) (APPEND (APPEND ys' [x]) zs))`--)
      
      THEN ASM_REWRITE_TAC[GSYM(APPEND_ASSOC)]       
      THEN SUBGOAL_THEN (--`LENGTH (ys':('a)list) <= (SUC n)`--) ASSUME_TAC
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= n`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN SUBGOAL_THEN (--`LENGTH (ys':('a)list) <= (SUC k)`--) ASSUME_TAC
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= k`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN IMP_RES_TAC (EL_APPEND2) THEN ASM_REWRITE_TAC[]
      THEN SUBGOAL_THEN (--`(SUC n) - (LENGTH (ys':('a)list)) = (SUC (n - (LENGTH (ys':('a)list))))`--) (fn thm => REWRITE_TAC[thm])
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= n`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN SUBGOAL_THEN (--`(SUC k) - (LENGTH (ys':('a)list)) = (SUC (k - (LENGTH (ys':('a)list))))`--) (fn thm => REWRITE_TAC[thm])
      THENL[UNDISCH_TAC (--`LENGTH (ys':('a)list) <= k`--) THEN RW_TAC arith_ss [],ALL_TAC]
      THEN REWRITE_TAC[EL, APPEND, TL]
      THEN STRIP_TAC 
      THEN FIRST_ASSUM MATCH_MP_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
      THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC  THEN UNDISCH_ONE_TAC 
      THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC 
      THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC 
      THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN RW_TAC arith_ss []
     ]
   ]
 ,ASM_REWRITE_TAC[]]
THEN CONJ_TAC
THENL
[ALL_TAC,
REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[IS_EL] THEN ASM_REWRITE_TAC[]]
THEN REWRITE_TAC[EXTENSION,IN_L2S_IS_EL] THEN GEN_TAC

THEN EQ_TAC THEN STRIP_TAC
THENL
 [SUBGOAL_THEN (--`~((x':'a) = x)`--) ASSUME_TAC
  THENL
   [IMP_RES_TAC NO_DUPLICATES_THM THEN STRIP_TAC THEN XRULE_ASSUM_TAC (ISPEC (--`x:'a`--))
    THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm,IS_EL,DEL_def])
    THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
   ,ALL_TAC] THEN XRULE_ASSUM_TAC (ISPEC (--`x':'a`--))
  THEN UNDISCH_TAC (--`IS_EL (x':'a) (CONS x xs) = IS_EL x' ys`--)
  THEN ASM_REWRITE_TAC[IS_EL , IS_EL_APPEND]
 ,
  SUBGOAL_THEN (--`~((x':'a) = x)`--) ASSUME_TAC
  THENL
   [IMP_RES_TAC NO_DUPLICATES_THM THEN STRIP_TAC 
    THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
    THEN UNDISCH_TAC (--`~(IS_EL (x:'a) (DEL x ys))`--) THEN ASM_REWRITE_TAC[]
   ,ALL_TAC] THEN XRULE_ASSUM_TAC (ISPEC (--`x':'a`--))
  THEN UNDISCH_TAC (--`IS_EL (x':'a) (CONS x xs) = IS_EL x' ys`--)
  THEN UNDISCH_TAC (--`IS_EL (x':'a) (APPEND ys' zs)`--) 
  THEN ASM_REWRITE_TAC[IS_EL , IS_EL_APPEND]
  THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] 
 ]);
 



val _ = export_theory();
