
(*----------------------------------------------------------------------*)
(* FILE: mk_pvt_justification.sml                                       *)
(* DESCRIPTION:                                                         *)
(*  The following new recursive type is added to the HOL logic          *)
(*  data type:							        *)
(*									*)
(*    pvt = NUM num							*)
(*        | SET (pvt)set 						*)
(*        | LIST (pvt)list						*)
(*        | TREE (pvt)ltree 						*)
(*									*)
(*  It is a conservative extension, in the sense that all theorems      *)
(*  needed to prove the abstract characterisation of pvt, the case      *)
(*  distinction, and the induction theorem are proved, and no axiom     *)
(*  is added.								*)
(*									*)
(*  The motivation for adding this type to the HOL logic, is to have a  *)
(*  justification for adding axioms to abstractly characterise the      *)
(*  following data type:						*)
(*									*)
(*	pvt = NUM num							*)
(*          | BOOL bool							*)
(*          | REAL real 						*)
(*          | STR string						*)
(*          | SET (pvt)set 						*)
(*          | LIST (pvt)list						*)
(*          | TREE (pvt)ltree 						*)
(*    									*)
(*                                                                      *) 
(*                                                                      *)
(* AUTHOR: Tanja Vos, Utrecht University                                *) 
(* E-Mail: tanja@cs.ruu.nl                                              *)
(* DATE: june 1998 							*)
(* REVISED HOL taupo 6: April 2002		                        *)
(*                                                                      *)
(*----------------------------------------------------------------------*)



open Parse HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ; 


val _ = new_theory "pvt_justification";

(*

load "bossLib";
load "more_ltreeTheory";
load "quotientTheory";
load "S2L_L2STheory";
load "setLib";
load "sumTheory";

*)

open bossLib;
open more_ltreeTheory;
open quotientTheory;
open S2L_L2STheory;
open setLib;
open setTheory;
open sumTheory;
open some_list_thmsTheory;
open rich_listTheory;
open listTheory;
open ltreeTheory;



val LIST_INDUCT_TAC = INDUCT_THEN list_INDUCT ASSUME_TAC;    
val UNDISCH_ONE_TAC =
    let
        fun th_tac (th:thm) (tac:tactic) = (UNDISCH_TAC (concl th)) THEN tac
        fun u_asm  (th:thm) = itlist  th_tac [th] ALL_TAC
    in
        FIRST_ASSUM u_asm
    end;
 
val XRULE_ASSUM_TAC = 
       let 
        fun app (rule:thm->thm) = (fn (t:thm) => (rule t) handle _ =>  t)
       in
       RULE_ASSUM_TAC o app
       end;
  
  
(*-------------------------------------------------*)
(* Some theorems about sum 		           *)
(*-------------------------------------------------*)
  


val sum_ISL = prove(
  (--`!(v:'a + 'b). (ISL v) ==> (?x. v = (INL x))`--),
   REPEAT STRIP_TAC THEN IMP_RES_TAC INL
   THEN EXISTS_TAC (--`OUTL (v:('a + 'b))`--)
   THEN ASM_REWRITE_TAC[]);
 
val sum_ISR = prove(
(--`!(v:'a + 'b). (ISR v) ==> (?(y:'b). v = (INR y))`--),
 REPEAT STRIP_TAC THEN IMP_RES_TAC INR
   THEN EXISTS_TAC (--`OUTR (v:('a + 'b))`--)
   THEN ASM_REWRITE_TAC[]);


val ISR_not_ISL = prove(
  (--`!(h:'a + 'b).
     (ISR h) = ~(ISL h)`--),
  STRIP_TAC THEN EQ_TAC
  THENL
   [STRIP_TAC THEN IMP_RES_TAC sum_ISR
    THEN ASM_REWRITE_TAC[ISL]
   ,
     REWRITE_TAC[IMP_DISJ_THM]
      THEN MATCH_ACCEPT_TAC ISL_OR_ISR
   ]);
   
val ISL_not_ISR = prove(
  (--`!(h:'a + 'b).
     (ISL h) = ~(ISR h)`--),
     REWRITE_TAC[ISR_not_ISL, NOT_CLAUSES]);
     


val INR_ONTO = prove(
  (--`!(x:'a) (y:'a) (v:('b + 'a)). ((v = (INR x)) /\ (v = (INR y))) ==> (x = y)`--),
  REPEAT STRIP_TAC
  THEN SUBGOAL_THEN (--`OUTR ((INR (x:'a)): ('b + 'a)) = OUTR ((INR (y:'a)):('b + 'a))`--) ASSUME_TAC
  THENL[XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] ,ALL_TAC]
  THEN UNDISCH_ONE_TAC
  THEN REWRITE_TAC[OUTR]);
  
val INR_ONTO2 = prove(
  (--`!(x:'a) (y:'a) .((INR x):('b + 'a) = (INR y)) ==> (x = y)`--),
   REPEAT STRIP_TAC
  THEN SUBGOAL_THEN (--`OUTR ((INR (x:'a)): ('b + 'a)) = OUTR ((INR (y:'a)):('b + 'a))`--) ASSUME_TAC
  THENL[XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] ,ALL_TAC]
  THEN UNDISCH_ONE_TAC
  THEN REWRITE_TAC[OUTR]);
    

(*-------------------------------------------------*)         
(* equivalence 				           *)
(* If we put sets in the first position of the +   *)
(*-------------------------------------------------*)


val plus = ty_antiq (==`:one + num + one + tree`==);

val equiv = new_definition ("equiv",
  (--`equiv = @fn. !(v:(^plus)) (tl:((^plus)ltree)list). 
                     fn (Node v tl) =
				       (\(t1:(^plus)ltree).
				           ((Tree_Root t1) = v)
				             /\
					     (((tl = (Tree_Subtrees t1)) /\ (?(n:num). v = (INR (INL n))))
					       \/
					       ((!t. (IS_EL t (Tree_Subtrees t1)) 
						        ==> (?t'. (IS_EL t' (MAP fn tl)) /\ (t' t)))
					         /\
						 (!t. (IS_EL t (MAP fn tl)) 
						        ==> (?t'. (IS_EL t' (Tree_Subtrees t1)) /\ (t t')))
					         /\ (ISL v))
						\/ 
						(((LENGTH tl) = (LENGTH (Tree_Subtrees t1))) 
						/\ (!n. (n < (LENGTH tl)) ==> ((EL n (MAP fn tl)) (EL n (Tree_Subtrees t1))))
						/\ ((v = (INR (INR (INL one)))) 
						     \/ 
						     (?(t:tree). (v = (INR (INR (INR t)))))) )
						 ))`--));
														
val equiv_THM = prove(
    (--`!(v:(^plus)) (tl:((^plus)ltree)list). 
        equiv (Node v tl) =  (\(t1:(^plus)ltree).
					          ((Tree_Root t1) = v)
						    /\
						    (((tl = (Tree_Subtrees t1)) /\ (?(n:num). v = (INR (INL n))))
					             \/
					             ((!t. (IS_EL t (Tree_Subtrees t1))
						        ==> (?t'. (IS_EL t' (MAP equiv tl)) /\ (t' t)))
					              /\
						      (!t. (IS_EL t (MAP equiv tl)) 
						        ==> (?t'. (IS_EL t' (Tree_Subtrees t1)) /\ (t t')))
						      /\ (ISL v))
						     \/
						     (((LENGTH tl) = (LENGTH (Tree_Subtrees t1))) 
						     /\ (!n. (n < (LENGTH tl)) ==> ((EL n (MAP equiv tl)) (EL n (Tree_Subtrees t1))))
						     /\ ((v = (INR (INR (INL one)))) 
						     \/ 
						     (?(t:tree). (v = (INR (INR (INR t)))))))
						     ))`--),
							
    REWRITE_TAC[equiv]  THEN CONV_TAC SELECT_CONV	 		
    THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:((^plus)ltree -> bool)list) (h:^plus) (tl:(((^plus)ltree)list)). 
                        (\(t1:(^plus)ltree). 
			    ((Tree_Root t1) = h)
			    /\
			    (((tl = (Tree_Subtrees t1)) /\ (?(n:num). h = (INR (INL n))))
			     \/
			    ((!t. (IS_EL t (Tree_Subtrees t1)) 
						        ==> (?t'. (IS_EL t' xs) /\ (t' t)))
			    /\
			    (!t. (IS_EL t xs) 
			         ==> (?t'. (IS_EL t' (Tree_Subtrees t1)) /\ (t t')))
			    /\ (ISL h))
			    \/
			    (((LENGTH tl) = (LENGTH (Tree_Subtrees t1))) 
			    /\ (!n. (n < (LENGTH tl)) ==> ((EL n xs) (EL n (Tree_Subtrees t1))))
			    /\ ((h = (INR (INR (INL one)))) 
						     \/ 
			     (?(t:tree). (h = (INR (INR (INR t))))))))))`--)
                   ltree_Axiom))) );


val equiv_lemma1 = prove(
    (--`!(v1:(^plus)) (tl1:((^plus)ltree)list) (v2:(^plus)) (tl2:((^plus)ltree)list). 
        equiv (Node v1 tl1)  (Node v2 tl2) =
					          ((v1 = v2)
						    /\
						    (((tl1 = tl2) /\ (?(n:num). v1 = (INR (INL n))))
					             \/
					             ((!t. (IS_EL t tl2)
						        ==> (?t'. (IS_EL t' (MAP equiv tl1)) /\ (t' t)))
					              /\
						      (!t. (IS_EL t (MAP equiv tl1)) 
						        ==> (?t'. (IS_EL t' tl2) /\ (t t')))
						      /\ (ISL v1))
						      \/
						      (((LENGTH tl1)  = (LENGTH tl2)) 
						      /\ (!n. (n < (LENGTH tl1)) ==> (equiv (EL n tl1) (EL n tl2)))
						      /\ ((v1 = (INR (INR (INL one)))) 
						         \/ 
						         (?(t:tree). (v1 = (INR (INR (INR t)))))) )))`--),
     REWRITE_TAC[equiv_THM] THEN BETA_TAC
     THEN REWRITE_TAC[Tree_Root, Tree_Subtrees] 
     THEN REPEAT GEN_TAC THEN EQ_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
     THENL[
     DISJ1_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
     ,
     DISJ2_TAC THEN DISJ2_TAC  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
     THEN REPEAT STRIP_TAC 
     THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}] 
                                       (EL_MAP))
     THEN RES_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
     ,
     DISJ2_TAC THEN DISJ2_TAC THEN CONJ_TAC
     THENL[ALL_TAC, DISJ2_TAC THEN EXISTS_TAC (--`t:tree`--) THEN ASM_REWRITE_TAC[]]
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
     THEN REPEAT STRIP_TAC 
     THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}] 
                                       (EL_MAP))
     THEN RES_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
     ,			       
     CONJ_TAC THENL[XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[],
     DISJ1_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]]
     ,
     XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN DISJ2_TAC THEN DISJ2_TAC 
     THEN REPEAT STRIP_TAC 
     THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}] 
                                       (EL_MAP))
     THEN ASM_REWRITE_TAC[] THEN RES_TAC 
     ,
     XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN DISJ2_TAC THEN DISJ2_TAC 
     THEN CONJ_TAC
     THENL[ALL_TAC, DISJ2_TAC THEN EXISTS_TAC (--`t:tree`--) THEN ASM_REWRITE_TAC[]]
     THEN REPEAT STRIP_TAC 
     THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}] 
                                       (EL_MAP))
     THEN RES_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
     
     ]);


val equiv_lemma2 = prove(
   (--`(!(t:(^plus)ltree). (IS_EL t tl2) ==> (?t'. (IS_EL t' tl1) /\ (equiv t' t)))
          =
	  (!(t:(^plus)ltree). (IS_EL t tl2) 
	         ==> (?(t':(^plus)ltree->bool). (IS_EL t' (MAP equiv tl1)) /\ (t' t)))`--),
   EQ_TAC  THEN REWRITE_TAC[GSYM IS_EL_MAP_EXISTS]
   THEN REPEAT STRIP_TAC THEN RES_TAC
   THENL
    [EXISTS_TAC (--`(equiv (t':(^plus)ltree)):(^plus)ltree->bool`--)
     THEN ASM_REWRITE_TAC[] 
     THEN  EXISTS_TAC (--`t':(^plus)ltree`--) THEN ASM_REWRITE_TAC[]
    ,
     EXISTS_TAC (--`x:(^plus)ltree`--) THEN XRULE_ASSUM_TAC SYM
     THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`(t':(^plus)ltree->bool) (t:(^plus)ltree)`--)
     THEN ASM_REWRITE_TAC[]]);
   
val equiv_lemma3 = prove(
   (--`(!(t:(^plus)ltree->bool). (IS_EL t (MAP equiv tl1)) ==> (?t'. (IS_EL  t' tl2) /\ t t'))
         =
	 (!t. (IS_EL t tl1) ==> (?t'. (IS_EL t' tl2) /\ (equiv t t')))`--),
  EQ_TAC  THEN REWRITE_TAC[GSYM IS_EL_MAP_EXISTS]
  THEN REPEAT STRIP_TAC THEN RES_TAC
   THENL
    [XRULE_ASSUM_TAC (ISPEC (--`equiv (t:(^plus)ltree):(^plus)ltree->bool`--))
    THEN FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC
    ,
     RES_TAC 
     THEN EXISTS_TAC (--`t'':(^plus)ltree`--) THEN XRULE_ASSUM_TAC SYM 
     THEN ASM_REWRITE_TAC[] 
    ] );
   


   
						     
val equiv_CHAR = prove(
   (--`!(v1:(^plus)) (tl1:((^plus)ltree)list) (v2:(^plus)) (tl2:((^plus)ltree)list). 
        equiv (Node v1 tl1)  (Node v2 tl2) =
					          ((v1 = v2)
						    /\
						    (((tl1 = tl2) /\ (?(n:num). v1 = (INR (INL n))))
					             \/
					             ((!t. (IS_EL t tl2)
						        ==> (?t'. (IS_EL t' tl1) /\ (equiv t' t)))
					              /\
						      (!t. (IS_EL t tl1) 
						        ==> (?t'. (IS_EL t' tl2) /\ (equiv t t')))
						      /\ (ISL v1))
						      \/
						      (((LENGTH tl1) = (LENGTH tl2)) 
						      /\ (!n. (n < (LENGTH tl1)) ==> (equiv (EL n tl1) (EL n tl2)))
						      /\ ((v1 = (INR (INR (INL one)))) 
						         \/ 
						        (?(t:tree). (v1 = (INR (INR (INR t)))))) )))`--),
REWRITE_TAC[GSYM equiv_lemma3,  equiv_lemma2,equiv_lemma1]);

  
val equiv_SYM = prove(
  (--`!(t1:(^plus)ltree). (\t1. !(t2:(^plus)ltree).  equiv t1 t2 ==> equiv t2 t1) t1`--),
  
REPEAT GEN_TAC THEN MATCH_MP_TAC (ltree_Induct)
THEN BETA_TAC THEN REWRITE_TAC[EVERY_IS_EL,equiv_CHAR]
THEN BETA_TAC THEN STRIP_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC
THEN STRIP_ASSUME_TAC (ISPEC (--`t2:(^plus)ltree`--) Node_onto)
THEN ASM_REWRITE_TAC[equiv_CHAR] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
THENL
 [XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,DISJ1_TAC THEN EXISTS_TAC (--`n:num`--) THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN DISJ2_TAC THEN DISJ1_TAC 
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  THEN CONJ_TAC
  THEN REPEAT STRIP_TAC THEN RES_TAC
  THENL
   [EXISTS_TAC (--`t'':(^plus)ltree`--) THEN ASM_REWRITE_TAC[]
   ,RES_TAC THEN EXISTS_TAC (--`t'':(^plus)ltree`--) THEN ASM_REWRITE_TAC[]
   ]
 ,XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,DISJ2_TAC THEN DISJ2_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  THEN REPEAT STRIP_TAC THEN RES_TAC THEN UNDISCH_ONE_TAC THEN FIRST_ASSUM MATCH_MP_TAC
  THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[]
 ,XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,DISJ2_TAC THEN DISJ2_TAC THEN CONJ_TAC
  THENL
  [XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC THEN RES_TAC
   THEN UNDISCH_ONE_TAC THEN FIRST_ASSUM MATCH_MP_TAC
   THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[]
  ,
   DISJ2_TAC THEN EXISTS_TAC (--`t':tree`--) THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  ]
]);


val equiv_TRANS = prove(
  (--`!(t1:(^plus)ltree). (\t1. !(t2:(^plus)ltree) (t3:(^plus)ltree).
         ((equiv t1 t2) /\ (equiv t2 t3)) ==> (equiv t1 t3)) t1`--),
         
REPEAT GEN_TAC THEN MATCH_MP_TAC ltree_Induct
THEN BETA_TAC THEN REWRITE_TAC[EVERY_IS_EL,equiv_CHAR]
THEN BETA_TAC THEN STRIP_TAC THEN STRIP_TAC THEN REPEAT GEN_TAC
THEN STRIP_ASSUME_TAC (ISPEC (--`t2:(^plus)ltree`--) Node_onto)
THEN STRIP_ASSUME_TAC (ISPEC (--`t3:(^plus)ltree`--) Node_onto)
THEN ASM_REWRITE_TAC[equiv_CHAR] THEN REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
THENL
 [DISJ1_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
 ,UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[ISL]
 ,UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
  THEN IMP_RES_TAC INR_ONTO2 
  THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n) : num + one + tree`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]
 ,
  UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
  THEN IMP_RES_TAC INR_ONTO2 
  THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n) : num + one + tree`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]
 ,DISJ2_TAC THEN DISJ1_TAC THEN CONJ_TAC THEN REPEAT STRIP_TAC 
  THEN RES_TAC THEN RES_TAC
   THENL
    [EXISTS_TAC (--`t''''':(^plus)ltree`--) THEN ASM_REWRITE_TAC[] THEN RES_TAC
    ,
     EXISTS_TAC (--`t''':(^plus)ltree`--) THEN ASM_REWRITE_TAC[] 
    ]
  ,UNDISCH_TAC (--`ISL (h:^plus)`--) THEN ASM_REWRITE_TAC[ISL]
  ,UNDISCH_TAC (--`ISL (h:^plus)`--) THEN ASM_REWRITE_TAC[ISL]
  ,UNDISCH_TAC (--`ISL (v:^plus)`--) THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
   THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[ISL]
  ,DISJ2_TAC THEN DISJ2_TAC THEN XRULE_ASSUM_TAC SYM THEN ONCE_ASM_REWRITE_TAC[]
   THEN REPEAT STRIP_TAC THEN RES_TAC THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--)
   THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN RES_TAC
   THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
   THENL[MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[],RES_TAC]
  ,UNDISCH_TAC (--`h:^plus = v`--)  THEN ASM_REWRITE_TAC[] 
   THEN STRIP_TAC THEN IMP_RES_TAC INR_ONTO2 THEN IMP_RES_TAC INR_ONTO2
   THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):one + tree`--) ISL_OR_ISR)
   THEN UNDISCH_ONE_TAC 
   THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]
  ,UNDISCH_TAC (--`h:^plus = v`--)  THEN ASM_REWRITE_TAC[] 
   THEN STRIP_TAC THEN IMP_RES_TAC INR_ONTO2 
   THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n):num + one + tree`--) ISL_OR_ISR)
   THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM
   THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]
  ,UNDISCH_TAC (--`ISL (v:^plus)`--) THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
   THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[ISL]
  ,UNDISCH_TAC (--`h:^plus = v`--)  THEN ASM_REWRITE_TAC[] 
   THEN STRIP_TAC THEN IMP_RES_TAC INR_ONTO2 THEN IMP_RES_TAC INR_ONTO2
   THEN STRIP_ASSUME_TAC (ISPEC (--`(INR t'):one + tree`--) ISL_OR_ISR)
   THEN UNDISCH_ONE_TAC 
   THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]
  ,DISJ2_TAC THEN DISJ2_TAC THEN CONJ_TAC
    THENL
     [XRULE_ASSUM_TAC SYM THEN ONCE_ASM_REWRITE_TAC[]
      THEN REPEAT STRIP_TAC THEN RES_TAC THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--)
      THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN RES_TAC
      THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
      THENL[MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[],RES_TAC]
     ,
      DISJ2_TAC THEN EXISTS_TAC (--`t':tree`--) THEN ASM_REWRITE_TAC[]
     ]
 ]);
     
 
val equiv_REFL = prove(
   (--`!(t:(^plus)ltree). (\t. equiv t t) t`--),
   
 
 STRIP_TAC THEN MATCH_MP_TAC ltree_Induct
 THEN BETA_TAC THEN REWRITE_TAC[EVERY_IS_EL,equiv_CHAR]
 THEN BETA_TAC THEN REPEAT STRIP_TAC
 
 THEN STRIP_ASSUME_TAC (ISPEC (--`h:^plus`--) ISL_OR_ISR) THEN ASM_REWRITE_TAC[]
 THEN IMP_RES_TAC sum_ISR 
 THENL
  [DISJ2_TAC THEN DISJ1_TAC THEN REPEAT STRIP_TAC THEN RES_TAC
   THEN EXISTS_TAC (--`t':(^plus)ltree`--) THEN ASM_REWRITE_TAC[] 
  ,
   
   STRIP_ASSUME_TAC (ISPEC (--`y:num + one + tree`--) ISL_OR_ISR) 
   THENL[IMP_RES_TAC sum_ISL 
         THEN DISJ1_TAC THEN EXISTS_TAC (--`x:num`--) THEN ASM_REWRITE_TAC[]
        ,IMP_RES_TAC (ISPEC (--`y:num + one + tree`--) sum_ISR)
        THEN STRIP_ASSUME_TAC (ISPEC (--`y': one + tree`--) ISL_OR_ISR) 
        THENL[IMP_RES_TAC sum_ISL THEN STRIP_ASSUME_TAC (ISPEC (--`x:one`--) oneTheory.one)
              THEN ASM_REWRITE_TAC[]  THEN DISJ2_TAC THEN DISJ2_TAC
              THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN MATCH_MP_TAC EL_IS_EL
              THEN ASM_REWRITE_TAC[]
             ,
              IMP_RES_TAC (ISPEC (--`y':one+tree`--) sum_ISR)
              THEN DISJ2_TAC THEN DISJ2_TAC THEN CONJ_TAC
                THENL
                 [REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN MATCH_MP_TAC EL_IS_EL
                  THEN ASM_REWRITE_TAC[]
                 ,
                  DISJ2_TAC THEN EXISTS_TAC (--`y'':tree`--) THEN ASM_REWRITE_TAC[]
                 ]
               
             ]
        ]
  ]);
  
  

val equiv_EQUIV_REL = store_thm("equiv_EQUIV_REL",
   (--`!(t1:(^plus)ltree) (t2:(^plus)ltree).
       (equiv t1 t2) = ((equiv t1) = (equiv t2))`--),
      
REWRITE_TAC[EQUIV_DEF, BETA_RULE equiv_SYM, BETA_RULE equiv_TRANS, BETA_RULE equiv_REFL]);
 
 
 

val equiv_lemma4 = prove(

  (--`!(tl1:((^plus)ltree)list) (tl2:((^plus)ltree)list).
     ((!t. (IS_EL t tl2) ==> (?t'. (IS_EL t' tl1) /\ (equiv t' t)))
      /\
      (!t. (IS_EL t tl1) ==> (?t'. (IS_EL t' tl2) /\ (equiv t t'))))
      =
      ((IMAGE equiv (L2S tl1)) = (IMAGE equiv (L2S tl2)))`--),
      
  REPEAT GEN_TAC 
  THEN REWRITE_TAC[EXTENSION,IN_IMAGE,IN_L2S_IS_EL]
  THEN EQ_TAC THEN REPEAT STRIP_TAC THEN RES_TAC
  THENL
   [EQ_TAC THEN REPEAT STRIP_TAC THEN RES_TAC
    THENL[EXISTS_TAC (--`t':(^plus)ltree`--) THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL],
          EXISTS_TAC (--`t':(^plus)ltree`--) THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL]
          THEN MATCH_MP_TAC (BETA_RULE equiv_SYM) THEN ASM_REWRITE_TAC[] ]
   ,
    ONCE_REWRITE_TAC[CONJ_SYM] THEN REWRITE_TAC[equiv_EQUIV_REL]
    THEN XRULE_ASSUM_TAC (ISPEC (--`equiv (t:(^plus)ltree)`--))
    THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV)
    THEN ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`t:(^plus)ltree`--) 
    THEN ASM_REWRITE_TAC[BETA_RULE equiv_REFL]
   ,
    ONCE_REWRITE_TAC[CONJ_SYM] THEN REWRITE_TAC[equiv_EQUIV_REL]
    THEN XRULE_ASSUM_TAC (ISPEC (--`equiv (t:(^plus)ltree)`--))
    THEN XRULE_ASSUM_TAC SYM
    THEN ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`t:(^plus)ltree`--) 
    THEN ASM_REWRITE_TAC[BETA_RULE equiv_REFL]
   ]);
   
 
 
val list_lemma = prove(
  (--`!(xs:('a)list) (ys:('a)list).
   (xs = ys) ==> (!n. (n < (LENGTH xs)) ==> ((EL n xs) = (EL n ys)     ))`--),
REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);


val equiv_lemma5 = prove(
      
   (--`!(tl1:((^plus)ltree)list) (tl2:((^plus)ltree)list).
     (((LENGTH tl1) = (LENGTH tl2)) /\ (!n. (n < (LENGTH tl1)) ==> (equiv (EL n tl1) (EL n tl2))))
        =
        ((MAP equiv tl1) = (MAP equiv tl2))`--),
        
  REPEAT GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
  THENL
   [MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] 
   THEN IMP_RES_TAC 
        (INST_TYPE [{redex = ==`:'b`== , residue = ==`:(one + num + one + tree)ltree->bool`==}] (EL_MAP))
   THEN ASM_REWRITE_TAC[]      
   THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC  THEN ASM_REWRITE_TAC[] 
   THEN STRIP_TAC
   THEN IMP_RES_TAC 
        (INST_TYPE [{redex = ==`:'b`== , residue = ==`:(one + num + one + tree)ltree->bool`==}] (EL_MAP))
   THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[GSYM equiv_EQUIV_REL] 
   THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[]
   ,
    REWRITE_TAC[GSYM (ISPEC (--`equiv:(^plus)ltree-> (^plus)ltree -> bool`--)
                     (ISPEC (--`tl1:((^plus)ltree)list`--) LENGTH_MAP))]
    THEN ASM_REWRITE_TAC[GSYM (ISPEC (--`equiv:(^plus)ltree-> (^plus)ltree -> bool`--)
                     (ISPEC (--`tl2:((^plus)ltree)list`--) LENGTH_MAP))]
   ,
    IMP_RES_TAC  list_lemma 
    THEN XRULE_ASSUM_TAC (REWRITE_RULE[LENGTH_MAP]) THEN RES_TAC
    THEN UNDISCH_ONE_TAC  
    THEN IMP_RES_TAC 
          (INST_TYPE [{redex = ==`:'b`== , residue = ==`:(one + num + one + tree)ltree->bool`==}]EL_MAP)
    THEN ONCE_ASM_REWRITE_TAC[] 
    THEN SUBGOAL_THEN (--`LENGTH (tl1:((^plus)ltree)list) = LENGTH (tl2:((^plus)ltree)list)`--) ASSUME_TAC
    THENL
     [REWRITE_TAC[GSYM (ISPEC (--`equiv:(^plus)ltree-> (^plus)ltree -> bool`--)
                     (ISPEC (--`tl1:((^plus)ltree)list`--) LENGTH_MAP))]
      THEN ASM_REWRITE_TAC[GSYM (ISPEC (--`equiv:(^plus)ltree-> (^plus)ltree -> bool`--)
                     (ISPEC (--`tl2:((^plus)ltree)list`--) LENGTH_MAP))]
     ,ALL_TAC]
     THEN UNDISCH_TAC (--`n < (LENGTH (tl1:((^plus)ltree)list))`--) THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
     THEN IMP_RES_TAC 
          (INST_TYPE [{redex = ==`:'b`== , residue = ==`:(one + num + one + tree)ltree->bool`==}]EL_MAP)
     THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[equiv_EQUIV_REL]
  ]);
  
     
   
 


val equiv_DEF = store_thm("equiv_DEF",
 (--`!(v1:(^plus)) (tl1:((^plus)ltree)list) (v2:(^plus)) (tl2:((^plus)ltree)list). 
        equiv (Node v1 tl1)  (Node v2 tl2) =
					          ((v1 = v2)
						    /\
						    (((tl1 = tl2) /\ (?(n:num). v1 = (INR (INL n))))
					             \/
					             (((IMAGE equiv (L2S tl1)) = (IMAGE equiv (L2S tl2))) /\ (ISL v1))
						      \/
						      (((MAP equiv tl1) = (MAP equiv tl2))
						      /\ ((v1 = (INR (INR (INL one)))) 
						         \/ 
						        (?(t:tree). (v1 = (INR (INR (INR t)))))) )))`--),
REPEAT GEN_TAC 
THEN REWRITE_TAC[GSYM equiv_lemma4,  GSYM equiv_lemma5]
THEN MATCH_ACCEPT_TAC (REWRITE_RULE [CONJ_ASSOC] equiv_CHAR));
 
       

val LIST_para_PRESERVE_part_TAC = 
   MATCH_MP_TAC LIST_EQ THEN ASM_REWRITE_TAC[LENGTH_MAP]
   THEN REPEAT STRIP_TAC 
   THEN IMP_RES_TAC (EL_MAP)
   THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
   THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM  THEN ASM_REWRITE_TAC[]
   THEN XRULE_ASSUM_TAC SYM THEN STRIP_TAC  THEN IMP_RES_TAC (EL_MAP)
   THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
   THEN STRIP_ASSUME_TAC (ISPEC (--`n:num`--) (arithmeticTheory.num_CASES))
   THENL
    [ASM_REWRITE_TAC[EL,HD] THEN RES_TAC
     THEN FIRST_ASSUM MATCH_MP_TAC
     THEN XRULE_ASSUM_TAC ((REWRITE_RULE[EL,HD]) o (ISPEC (--`0`--)))
     THEN FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[LENGTH]
     THEN DECIDE_TAC
    ,
     ASM_REWRITE_TAC[EL,TL] 
    THEN XRULE_ASSUM_TAC (BETA_RULE o REWRITE_RULE [EVERY_IS_EL])
    THEN XRULE_ASSUM_TAC ((REWRITE_RULE[EL,TL]) o (ISPEC (--`SUC n'`--)))
    THEN SUBGOAL_THEN (--`IS_EL (EL n' (t:((^plus)ltree)list)) t`--) ASSUME_TAC
     THENL[MATCH_MP_TAC EL_IS_EL
      THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--) 
      THEN ASM_REWRITE_TAC[LENGTH] THEN UNDISCH_ONE_TAC
      THEN DECIDE_TAC,ALL_TAC]
      THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--) 
      THEN ASM_REWRITE_TAC[LENGTH] THEN DECIDE_TAC
    ];
    


(*-------------------------------------------------*)
(* Theorem stating the conditions under which      *)
(* recursive functions                             *)
(* on ltree (i.e. ltree_Axiom) preserve equiv.     *)
(*-------------------------------------------------*)


val SET_EQ1_TAC = 
 MATCH_MP_TAC EL_IMP_IS_EL THEN IMP_RES_TAC IS_EL_IMP_EL
 THEN SUBGOAL_THEN (--`(x = (EL n (MAP (fn:(^plus)ltree->'b) (CONS (h:(^plus)ltree) t)), 
                              EL n (MAP (equiv:(^plus)ltree->(^plus)ltree->bool) (CONS h t)))) 
                              /\ (n < (LENGTH (MAP fn (CONS h t))))`--) ASSUME_TAC
 THENL
  [XRULE_ASSUM_TAC (REWRITE_RULE [ZIP_MAP_LENGTH])
   THEN ASM_REWRITE_TAC[] THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[LENGTH_MAP]
    ,
   ALL_TAC]
  THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN SUBGOAL_THEN (--`(n < (LENGTH (CONS (h:(^plus)ltree) t)))`--) ASSUME_TAC
  THENL[UNDISCH_ONE_TAC THEN REWRITE_TAC[ZIP_MAP_LENGTH,LENGTH_MAP], ALL_TAC]
  THEN IMP_RES_TAC EL_MAP THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}]
                 EL_MAP)
  THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # ((^plus)ltree->bool))`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) THEN STRIP_TAC
  THEN SUBGOAL_THEN (--`?(x':(^plus)ltree). (EL n (CONS h t)) = x'`--) STRIP_ASSUME_TAC
  THENL[EXISTS_TAC (--`(EL n (CONS (h:(^plus)ltree) t))`--) THEN REFL_TAC,ALL_TAC]
  THEN XRULE_ASSUM_TAC SYM 
  THEN IMP_RES_TAC EL_IMP_IS_EL THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN STRIP_TAC
  THEN XRULE_ASSUM_TAC SYM 
  THEN RES_TAC THEN IMP_RES_TAC (ISPEC (--`trl:((^plus)ltree)list`--)
                                (ISPEC (--`t':(^plus)ltree`--) IS_EL_IMP_EL))
  THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN EXISTS_TAC (--`n':num`--) THEN ASM_REWRITE_TAC[ZIP_MAP_LENGTH,LENGTH_MAP]
  THEN SUBGOAL_THEN (--`(EL n' (ZIP (MAP (fn:(^plus)ltree->'b) trl,
                                    MAP (equiv:(^plus)ltree->(^plus)ltree->bool) trl))) =
                        (EL n' (MAP fn trl), EL n' (MAP equiv trl))`--) ASSUME_TAC
  THENL
   [MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
   ,ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
   ]
  THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # ((^plus)ltree->bool))`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN CONJ_TAC
  THENL
   [(* FST *)
    ASM_REWRITE_TAC[]
    THEN IMP_RES_TAC (ISPEC (--`trl:((^plus)ltree)list`--) (SPEC (--`n':num`--) EL_MAP))
    THEN ASM_REWRITE_TAC[] 
    THEN ASM_CASES_TAC (--`x' = (h:(^plus)ltree)`--)
    THENL
     [ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_TAC (--`(equiv:(^plus)ltree -> (^plus)ltree -> bool) x' t'`--)
      THEN ASM_REWRITE_TAC[]
     ,
      SUBGOAL_THEN (--`IS_EL (x':(^plus)ltree) t`--) ASSUME_TAC
      THENL[UNDISCH_TAC (--`IS_EL (x':(^plus)ltree) (CONS h t)`--)
            THEN ASM_REWRITE_TAC[IS_EL],ALL_TAC]
      THEN RES_TAC THEN ASM_REWRITE_TAC[]
     ]
   ,
    (* SND *)
    
    
    
    IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}]
                 EL_MAP)
    THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
    THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC)
    THEN UNDISCH_TAC (--`(equiv:(^plus)ltree -> (^plus)ltree -> bool) x' t'`--)
    THEN ASM_REWRITE_TAC[equiv_EQUIV_REL]
   ];
   
 
  


val SET_EQ2_TAC = 

 MATCH_MP_TAC EL_IMP_IS_EL THEN IMP_RES_TAC IS_EL_IMP_EL
 THEN SUBGOAL_THEN (--`(x = (EL n (MAP (fn:(^plus)ltree->'b) trl), 
                              EL n (MAP (equiv:(^plus)ltree->(^plus)ltree->bool) trl))) 
                              /\ (n < (LENGTH (MAP fn trl)))`--) ASSUME_TAC
 THENL
  [XRULE_ASSUM_TAC (REWRITE_RULE [ZIP_MAP_LENGTH])
   THEN ASM_REWRITE_TAC[] THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[LENGTH_MAP]
    ,
   ALL_TAC]
  THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN SUBGOAL_THEN (--`(n < (LENGTH (trl:((^plus)ltree)list)))`--) ASSUME_TAC
  THENL[UNDISCH_ONE_TAC THEN REWRITE_TAC[ZIP_MAP_LENGTH,LENGTH_MAP], ALL_TAC]
  THEN IMP_RES_TAC EL_MAP THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}]
                 EL_MAP)
  THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # ((^plus)ltree->bool))`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) THEN STRIP_TAC
  THEN SUBGOAL_THEN (--`?(x':(^plus)ltree). (EL n trl) = x'`--) STRIP_ASSUME_TAC
  THENL[EXISTS_TAC (--`(EL n (trl:((^plus)ltree)list))`--) THEN REFL_TAC,ALL_TAC]
  THEN XRULE_ASSUM_TAC SYM 
  THEN IMP_RES_TAC EL_IMP_IS_EL THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN UNDISCH_ONE_TAC THEN POP_ASSUM (fn thm => ALL_TAC) THEN STRIP_TAC
  THEN XRULE_ASSUM_TAC SYM 
  THEN RES_TAC THEN IMP_RES_TAC (ISPEC (--`(CONS h t):((^plus)ltree)list`--)
                                (ISPEC (--`t':(^plus)ltree`--) IS_EL_IMP_EL))
                                
  THEN EXISTS_TAC (--`n':num`--) THEN ASM_REWRITE_TAC[ZIP_MAP_LENGTH,LENGTH_MAP]
  THEN SUBGOAL_THEN (--`(EL n' (ZIP (MAP (fn:(^plus)ltree->'b) (CONS h t),
                                    MAP (equiv:(^plus)ltree->(^plus)ltree->bool) (CONS h t)))) =
                        (EL n' (MAP fn (CONS h t)), EL n' (MAP equiv (CONS h t)))`--) ASSUME_TAC
  THENL
   [MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
   ,ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
   ]
  THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # ((^plus)ltree->bool))`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN CONJ_TAC
  THENL
   [(* FST *)
    ASM_REWRITE_TAC[]
    THEN IMP_RES_TAC (ISPEC (--`(CONS h t):((^plus)ltree)list`--) (SPEC (--`n':num`--) EL_MAP))
    THEN ASM_REWRITE_TAC[] 
    THEN STRIP_ASSUME_TAC (ISPEC (--`n':num`--) (arithmeticTheory.num_CASES))
    THENL
     [ASM_REWRITE_TAC[EL,HD] THEN CONV_TAC SYM_CONV
     THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_TAC (--`(equiv:(^plus)ltree -> (^plus)ltree -> bool) t' x'`--)
      THEN ASM_REWRITE_TAC[EL,HD]
     ,
      ASM_REWRITE_TAC[EL,TL]
      THEN SUBGOAL_THEN (--`(IS_EL (EL n'' (t:((^plus)ltree)list)) t)`--) ASSUME_TAC
      THENL[MATCH_MP_TAC EL_IS_EL THEN UNDISCH_TAC (--`n' < LENGTH (CONS (h:(^plus)ltree) t)`--)
            THEN ASM_REWRITE_TAC[LENGTH] THEN DECIDE_TAC
      ,ALL_TAC]
      THEN UNDISCH_TAC (--`t' = EL n' (CONS (h:(^plus)ltree) t)`--)
      THEN ASM_REWRITE_TAC[EL,TL] THEN STRIP_TAC THEN CONV_TAC SYM_CONV
      THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC
      THEN UNDISCH_TAC (--`(equiv:(^plus)ltree -> (^plus)ltree -> bool) t' x'`--)
      THEN ASM_REWRITE_TAC[]
     ]
   ,
    (* SND *)
    IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree -> bool`==}]
                 EL_MAP)
    THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) 
    THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC)
    THEN UNDISCH_TAC (--`(equiv:(^plus)ltree -> (^plus)ltree -> bool) t' x'`--)
    THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
    THEN CONV_TAC SYM_CONV THEN REWRITE_TAC[GSYM equiv_EQUIV_REL] THEN ASM_REWRITE_TAC[] 
   ];
   
 

val ltree_Axiom_PRESERVES_equiv_lemma = prove(
(--`!(t1:(^plus)ltree).
         (\t1. !(t2:(^plus)ltree) (fn:(^plus)ltree->'b) (f:('b)list->(^plus) ->((^plus)ltree)list->'b).
	           ((equiv t1 t2) /\ (!v tl. fn (Node v tl) = f (MAP fn tl) v tl)
		   /\
		   (!(xs1:('b)list) (xs2:('b)list) (tl1:((^plus)ltree)list) (tl2:((^plus)ltree)list) (v:^plus).
		      ((equiv (Node v tl1) (Node v tl2))
		     /\ ((ISL v) ==> (((LENGTH xs1) = (LENGTH tl1)) /\ ((LENGTH xs2) = (LENGTH tl2)) /\ 
		          ((L2S (ZIP(xs1,(MAP equiv tl1)))) = (L2S (ZIP(xs2,(MAP equiv tl2)))))))
		     /\ ((ISR v) ==> (xs1 = xs2)))
		      ==> ((f xs1 v tl1) = (f xs2 v tl2))))
		   ==>
		   ((fn t1) = (fn t2))) t1`--),
		
 
MATCH_MP_TAC ltree_Induct
THEN LIST_INDUCT_TAC
THEN REWRITE_TAC[EVERY_DEF] 
THENL
 [BETA_TAC THEN REPEAT GEN_TAC
  THEN STRIP_ASSUME_TAC (ISPEC (--`t2:(^plus)ltree`--) Node_onto)
  THEN ASM_REWRITE_TAC[equiv_CHAR,MAP,LENGTH]
  THEN REWRITE_TAC[IS_EL,not_IS_EL_EQ_EMPTY]
  THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[MAP]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[MAP]
  THEN XRULE_ASSUM_TAC (REWRITE_RULE[LENGTH_NIL]) 
  THEN ASM_REWRITE_TAC[MAP] 
  
  
 ,
  BETA_TAC  THEN POP_ASSUM (fn thm =>  ALL_TAC)
  THEN GEN_TAC THEN REWRITE_TAC[EVERY_IS_EL] THEN BETA_TAC
  THEN STRIP_TAC THEN REPEAT GEN_TAC
  THEN STRIP_ASSUME_TAC (ISPEC (--`t2:(^plus)ltree`--) Node_onto)
  THEN ASM_REWRITE_TAC[] 
  THEN REPEAT STRIP_TAC
  THEN UNDISCH_TAC (--`equiv (Node (h':^plus) (CONS h (t:((^plus)ltree)list))) 
                                            (Node v (trl:((^plus)ltree)list))`--)
  THEN REWRITE_TAC[equiv_CHAR, NOT_CONS_NIL] 
  THEN STRIP_TAC
  THENL
  [(* NUM *)
   UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
  ,(* SET *)
  ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC 
  THEN XRULE_ASSUM_TAC SYM
  THEN ASM_REWRITE_TAC[equiv_CHAR,ISR_not_ISL,LENGTH_MAP] 
  THEN REWRITE_TAC[EXTENSION,L2S_MAP_eq_IMAGE] THEN STRIP_TAC
  THEN REWRITE_TAC[IN_IMAGE,IN_L2S_IS_EL]
  THEN EQ_TAC THEN STRIP_TAC 
  THENL
  [SET_EQ1_TAC, SET_EQ2_TAC]
 ,
  (* LIST *) 
   XRULE_ASSUM_TAC SYM 
   THEN ASM_REWRITE_TAC[]
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN ASM_REWRITE_TAC[equiv_CHAR]
   THEN XRULE_ASSUM_TAC SYM  THEN ASM_REWRITE_TAC[ISR,ISL]
   THEN XRULE_ASSUM_TAC SYM
   THEN LIST_para_PRESERVE_part_TAC
  ,
  (* TREE *) 
   XRULE_ASSUM_TAC SYM 
   THEN ASM_REWRITE_TAC[]
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN CONJ_TAC
   THENL[ASM_REWRITE_TAC[equiv_CHAR] THEN DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC 
   THEN EXISTS_TAC (--`t':tree`--) THEN ASM_REWRITE_TAC[],ALL_TAC]
   THEN XRULE_ASSUM_TAC SYM  THEN ASM_REWRITE_TAC[ISR,ISL]
   THEN XRULE_ASSUM_TAC SYM THEN LIST_para_PRESERVE_part_TAC
   
 ] 
]);


val ltree_Axiom_PRESERVES_equiv = store_thm("ltree_Axiom_PRESERVES_equiv",
(--`!(t1:(^plus)ltree) (t2:(^plus)ltree) (fn:(^plus)ltree->'b) (f:('b)list->^plus ->((^plus)ltree)list->'b).
	           ((equiv t1 t2) /\ (!v tl. fn (Node v tl) = f (MAP fn tl) v tl)
		   /\
		   (!(xs1:('b)list) (xs2:('b)list) (tl1:((^plus)ltree)list) (tl2:((^plus)ltree)list) (v:^plus).
		      ((equiv (Node v tl1) (Node v tl2)) /\ 
		      ((ISL v) ==> (((LENGTH xs1) = (LENGTH tl1)) /\ ((LENGTH xs2) = (LENGTH tl2)) 
		       /\ ((L2S (ZIP(xs1,(MAP equiv tl1)))) = (L2S (ZIP(xs2,(MAP equiv tl2))))))) /\
		      ((ISR v) ==> (xs1 = xs2)))
		      ==> ((f xs1 v tl1) = (f xs2 v tl2))))
		   ==>
		   ((fn t1) = (fn t2))`--),
  MATCH_ACCEPT_TAC (BETA_RULE ltree_Axiom_PRESERVES_equiv_lemma));
 

(*-------------------------------------------------*)
(* The new type 				   *)
(*-------------------------------------------------*)


val R_DEF = new_definition("R_DEF",
  (--`R (lt:(^plus)ltree) = ((?n. ((Tree_Root lt) = (INR (INL n)))) ==> ((Tree_Subtrees lt) = []))
                           /\
                           ((?(t:tree). ((Tree_Root lt) = (INR (INR (INR t))))) 
                                    ==> (Is_ltree (OUTR(OUTR(OUTR(Tree_Root lt))), (Tree_Subtrees lt))))`--));
                                    

val id = new_definition ("id",
  (--`id (x:'a) = x`--));  

val Q_HILB = new_definition("Q_HILB",
   (--`Q = @(fn:(^plus)ltree -> bool). 
          (!(v:^plus) (tl:((^plus)ltree)list). (fn (Node v tl) = ((R (Node v tl)) /\ (EVERY id (MAP fn tl)))))`--));   
         

val Q_THM = prove(
  (--`!(v:^plus) (tl:((^plus)ltree)list). 
      (Q (Node v tl)) = ((R (Node v tl)) /\ (EVERY id (MAP Q tl)))`--),
  
  REWRITE_TAC[Q_HILB] THEN CONV_TAC SELECT_CONV
  THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:(bool)list) (h:^plus) (tl:(((^plus)ltree)list)). ((R (Node h tl)) /\ (EVERY id xs)))`--)
                   ltree_Axiom))) );

val Q_DEF = store_thm("Q_DEF",
  (--`!(v:^plus) (tl:((^plus)ltree)list). (Q (Node v tl)) = (R (Node v tl)) /\ (EVERY Q tl)`--),
  
  REWRITE_TAC[Q_THM,EVERY_IS_EL, GSYM IS_EL_MAP_EXISTS] THEN REPEAT GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
  THEN ASM_REWRITE_TAC[] THEN RES_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[id]
  THEN XRULE_ASSUM_TAC (ISPEC (--`Q (x:(^plus)ltree)`--)) THEN UNDISCH_ONE_TAC
  THEN REWRITE_TAC[id]);
 
  


val Is_pvt_REP = new_definition ("Is_pvt_REP",
  (--`(Is_pvt_REP:((^plus)ltree->bool)-> bool) = (Q // equiv)`--));
     	 

val not_empty_thm = prove(
  (--`?(s:(^plus)ltree->bool). (Is_pvt_REP s)`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def] THEN BETA_TAC
  THEN EXISTS_TAC (--`equiv ((Node (INL one) []):(^plus)ltree)`--)
  THEN EXISTS_TAC (--`(Node (INL one) []):(^plus)ltree`--) 
  THEN REWRITE_TAC[R_DEF,Q_DEF, equiv_EQUIV_REL, equiv_CHAR,Tree_Root,Tree_Subtrees,EVERY_DEF]
  THEN STRIP_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):^plus`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THENL[ASM_REWRITE_TAC[], ALL_TAC] THEN REWRITE_TAC[ISL,ISR]);
  

val Is_pvt_REP_lemma = prove(
  (--`!(x:(^plus)ltree). (Q x) ==> Is_pvt_REP (equiv x)`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def] THEN BETA_TAC
  THEN REPEAT STRIP_TAC THEN EXISTS_TAC (--`x:(^plus)ltree`--)
  THEN ASM_REWRITE_TAC[]);
  

(* The type definition *)

val pvt_TY_DEF =
    new_type_definition{name = "pvt", 
                        pred = (--`Is_pvt_REP`--), 
                        inhab_thm = not_empty_thm};  

val pvt_ISO_DEF =
   define_new_type_bijections 
    {name = "pvt_ISO_DEF", 
     ABS = "ABS_pvt", 
     REP = "REP_pvt" , 
     tyax = pvt_TY_DEF};
     


(* The constructors *)


val pick = new_definition ("pick",
  (--`pick (P:('a)ltree -> bool) = @t. P t`--));
  



val NUM_DEF = new_definition("NUM_DEF",
   (--`NUM (n:num) = ABS_pvt (equiv (Node (INR (INL n)) []))`--));


val SET_DEF = new_definition("SET_DEF",
   (--`SET (s:(pvt)set) =  ABS_pvt(equiv (Node (INL one) (MAP (pick o REP_pvt) (S2L s))))`--));

val LIST_DEF = new_definition ("LIST_DEF",
  (--`LIST (l:(pvt)list) = ABS_pvt(equiv (Node (INR (INR (INL one))) (MAP (pick o REP_pvt) l)))`--));
  

val TREE_DEF = new_definition("TREE_DEF",
   (--`TREE (t:(pvt)ltree) = ABS_pvt(equiv (Node (INR (INR (INR (FST (REP_ltree t)))))
                                                                            (MAP  (pick o REP_pvt) (SND (REP_ltree t)))))`--));



val REP_ONTO = prove_rep_fn_onto pvt_ISO_DEF;
val ABS_ONTO = prove_abs_fn_onto pvt_ISO_DEF;
val REP_ONE = prove_rep_fn_one_one pvt_ISO_DEF;
val ABS_ONE = prove_abs_fn_one_one pvt_ISO_DEF;


val REP_pvt_EQ_equiv = prove(
(--`!(x:pvt). ?(t:(^plus)ltree). ((REP_pvt x) = (equiv t)) /\ (Q t)`--),

STRIP_TAC 
THEN REWRITE_TAC[ISPEC (--`REP_pvt (x:pvt)`--) 
((BETA_RULE o (REWRITE_RULE [Is_pvt_REP,QUOTIENT_SET_def])) REP_ONTO)]
THEN EXISTS_TAC (--`x:pvt`--) THEN REFL_TAC);

   


val EXISTS_REP_lemma = prove(
(--`!(x:pvt). ?(r:(^plus)ltree->bool). (x = (ABS_pvt r)) /\ (r = (REP_pvt x)) /\ (Is_pvt_REP r)`--),

STRIP_TAC  
THEN EXISTS_TAC (--`REP_pvt (x:pvt)`--)
THEN REWRITE_TAC[pvt_ISO_DEF]);



val equiv_pick_equiv = prove(
  (--`!(t:(^plus)ltree). 
      (equiv t) = (equiv (pick (equiv t)))`--),
     REWRITE_TAC[pick, equiv_EQUIV_REL]
     THEN STRIP_TAC THEN CONV_TAC SELECT_CONV
     THEN EXISTS_TAC (--`t:(^plus)ltree`--) THEN REFL_TAC);


val equiv_pick_REP_pvt = store_thm("equiv_pick_REP_pvt",
(--`!(x:pvt). 
       equiv (pick (REP_pvt x))  =  (REP_pvt x)`--),
  
GEN_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`x:pvt`--) REP_pvt_EQ_equiv)
THEN ASM_REWRITE_TAC[] THEN CONV_TAC SYM_CONV
THEN MATCH_ACCEPT_TAC equiv_pick_equiv);


val R_INR_INL = prove(
 (--`!(n:num) (t:((^plus)ltree)list). (R (Node (INR (INL n)) t)) ==> (t = [])`--),
 REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR] THEN REPEAT STRIP_TAC
 THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]);


val Is_pvt_REP_equiv_INR_INL = prove(
  (--`!(n:num). (Is_pvt_REP (equiv (Node (INR (INL n)) [])))`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def,Q_DEF,Tree_Root,Tree_Subtrees] THEN BETA_TAC
  THEN STRIP_TAC THEN EXISTS_TAC (--`Node ((INR (INL n)):^plus) ([]:((^plus)ltree)list)`--)
  THEN REWRITE_TAC[equiv_EQUIV_REL, equiv_CHAR,Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,EVERY_DEF]
  THEN STRIP_TAC THEN IMP_RES_TAC INR_ONTO2 THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n) :num + one + tree`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]);
  

 
val R_INR_INR_INR = prove(
   (--`!(t:tree) (tl:((^plus)ltree)list). (R (Node (INR (INR (INR t))) tl)) ==> (Is_ltree(t, tl))`--),
 REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR] THEN REPEAT STRIP_TAC
 THEN FIRST_ASSUM MATCH_MP_TAC THEN  EXISTS_TAC (--`t:tree`--) THEN ASM_REWRITE_TAC[]);


val Is_pvt_REP_equiv_INR_INR_INR = prove(
  (--`!(t:tree) (tl:((^plus)ltree)list).  ((Is_ltree(t, tl)) /\ (EVERY Q tl))
            ==> (Is_pvt_REP (equiv ((Node (INR (INR (INR t))) tl))))`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def,Q_DEF,Tree_Root,Tree_Subtrees] THEN BETA_TAC
  THEN REPEAT STRIP_TAC THEN EXISTS_TAC (--`Node ((INR (INR (INR t))):^plus) (tl:((^plus)ltree)list)`--)
  THEN REWRITE_TAC[equiv_EQUIV_REL, equiv_CHAR,Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,EVERY_DEF,OUTR]
  THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC 
  THEN IMP_RES_TAC INR_ONTO2 THEN STRIP_ASSUME_TAC (ISPEC (--`(INR (INR t)) :num + one + tree`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]);
  



val R_INL = prove(
   (--`!(tl:((^plus)ltree)list). (R (Node (INL one) tl))`--),
  REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR] THEN REPEAT STRIP_TAC 
  THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):^plus`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[]
  THEN REWRITE_TAC[ISL]);

val Is_pvt_REP_equiv_INL = prove(
  (--`!(tl:((^plus)ltree)list). (EVERY Q tl) ==> (Is_pvt_REP (equiv (Node (INL one) tl)))`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def,Q_DEF,Tree_Root,Tree_Subtrees] THEN BETA_TAC
  THEN REPEAT STRIP_TAC THEN EXISTS_TAC (--`Node ((INL one):^plus) (tl:((^plus)ltree)list)`--)
  THEN REWRITE_TAC[equiv_EQUIV_REL, equiv_CHAR,Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,EVERY_DEF,OUTR]
  THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
  THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one) :^plus`--) ISL_OR_ISR)
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]);
  
  


val R_INR_INR_INL = prove(
   (--`!(tl:((^plus)ltree)list). (R (Node (INR (INR (INL one))) tl))`--),  
  REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR] THEN REPEAT STRIP_TAC 
  THEN IMP_RES_TAC INR_ONTO2
  THENL
   [STRIP_ASSUME_TAC (ISPEC (--`(INR (INL one)):num + one + tree`--) ISL_OR_ISR)
    THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC[ISR]
   ,
    IMP_RES_TAC INR_ONTO2
    THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):one + tree`--) ISL_OR_ISR)
    THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC[ISL]
   ]);
   

val Is_pvt_REP_equiv_INR_INR_INL = prove(
  (--`!(tl:((^plus)ltree)list). (EVERY Q tl) ==> (Is_pvt_REP (equiv (Node (INR (INR (INL one))) tl)))`--),
  REWRITE_TAC[Is_pvt_REP,QUOTIENT_SET_def,Q_DEF,Tree_Root,Tree_Subtrees] THEN BETA_TAC
  THEN REPEAT STRIP_TAC THEN EXISTS_TAC (--`Node ((INR (INR (INL one))):^plus) (tl:((^plus)ltree)list)`--)
  THEN REWRITE_TAC[equiv_EQUIV_REL, equiv_CHAR,Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,EVERY_DEF,OUTR]
  THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC THEN IMP_RES_TAC INR_ONTO2
  THENL
   [STRIP_ASSUME_TAC (ISPEC (--`(INR (INL one)) :num + one + tree`--) ISL_OR_ISR)
    THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]
   ,
    IMP_RES_TAC INR_ONTO2
    THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one) :one + tree`--) ISL_OR_ISR)
    THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]
   ]);
   
  

val Q_IMP_equiv_Q = prove(
  (--`!(x:(^plus)ltree). (\x. (Q x) ==> (!t. (equiv t x) ==> (Q t))) x`--),
 MATCH_MP_TAC ltree_Induct THEN BETA_TAC
 THEN REWRITE_TAC[EVERY_IS_EL] THEN BETA_TAC THEN GEN_TAC THEN STRIP_TAC
 THEN GEN_TAC THEN REWRITE_TAC[Q_DEF,EVERY_IS_EL] THEN STRIP_TAC THEN GEN_TAC
 THEN STRIP_ASSUME_TAC (ISPEC (--`t':(^plus)ltree`--) Node_onto)
 THEN ASM_REWRITE_TAC[equiv_CHAR] THEN REPEAT STRIP_TAC THEN REWRITE_TAC[Q_DEF,EVERY_IS_EL]
 THEN CONJ_TAC
 THENL
  [POP_ASSUM (fn thm => ALL_TAC) THEN ASM_REWRITE_TAC[]
  ,
   ASM_REWRITE_TAC[]
  ,
   IMP_RES_TAC sum_ISL THEN STRIP_ASSUME_TAC (ISPEC (--`x:one`--) (oneTheory.one))
   THEN ASM_REWRITE_TAC[R_INL]
  ,
   REPEAT STRIP_TAC THEN RES_TAC THEN RES_TAC
  ,
   ASM_REWRITE_TAC[R_INR_INR_INL]
  ,
   REPEAT STRIP_TAC 
   THEN IMP_RES_TAC IS_EL_IMP_EL THEN ASM_REWRITE_TAC[] THEN RES_TAC
   THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
   THENL[MATCH_MP_TAC EL_IS_EL THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[],ALL_TAC]
   THEN RES_TAC
  ,
   UNDISCH_TAC (--`R (Node (h:^plus) t)`--) THEN UNDISCH_ONE_TAC
   THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
   THEN REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR,Is_ltree]
   THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
   THENL
    [IMP_RES_TAC INR_ONTO2 THEN STRIP_ASSUME_TAC (ISPEC (--`(INR(INR (t'':tree))):num + one + tree`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]
    ,
     FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`t'':tree`--) THEN REFL_TAC
    ]
   
  ,
   REPEAT STRIP_TAC 
   THEN IMP_RES_TAC IS_EL_IMP_EL THEN ASM_REWRITE_TAC[] THEN RES_TAC
   THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
   THENL[MATCH_MP_TAC EL_IS_EL THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[],ALL_TAC]
   THEN RES_TAC
  ]);
  

val Q_IMP_Q_pick_equiv = prove(
  (--`!(t:(^plus)ltree). (\t. (Q t) ==> (Q (pick (equiv t)))) t`--),
 MATCH_MP_TAC ltree_Induct THEN BETA_TAC
 THEN REWRITE_TAC[EVERY_IS_EL] THEN BETA_TAC THEN REPEAT STRIP_TAC
 THEN STRIP_ASSUME_TAC (ISPEC (--`(Node (h:^plus) t)`--) (REWRITE_RULE [GSYM equiv_EQUIV_REL] equiv_pick_equiv))
 THEN UNDISCH_ONE_TAC
 THEN STRIP_ASSUME_TAC (ISPEC (--`pick (equiv (Node (h:^plus) t))`--) Node_onto)
 THEN ASM_REWRITE_TAC[equiv_CHAR] THEN XRULE_ASSUM_TAC SYM THEN REPEAT STRIP_TAC
 THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 THEN REWRITE_TAC[Q_DEF] 
 THEN UNDISCH_TAC (--`Q (Node (h:^plus) t)`--) 
 THEN REWRITE_TAC[Q_DEF ,EVERY_IS_EL] THEN REPEAT STRIP_TAC THEN RES_TAC
 THENL
  [IMP_RES_TAC sum_ISL THEN STRIP_ASSUME_TAC (ISPEC (--`x:one`--) oneTheory.one)
   THEN ASM_REWRITE_TAC[R_INL]
  ,
   RES_TAC THEN IMP_RES_TAC (REWRITE_RULE[equiv_EQUIV_REL] (BETA_RULE Q_IMP_equiv_Q)) 
   THEN FIRST_ASSUM MATCH_MP_TAC THEN CONV_TAC SYM_CONV
   THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL] 
  ,
   XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[R_INR_INR_INL]
  ,
   IMP_RES_TAC IS_EL_IMP_EL THEN ASM_REWRITE_TAC[] THEN RES_TAC
   THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
   THENL[MATCH_MP_TAC EL_IS_EL THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[],ALL_TAC]
   THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--) THEN ASM_REWRITE_TAC[]
   THEN STRIP_TAC THEN RES_TAC
   THEN IMP_RES_TAC (REWRITE_RULE[equiv_EQUIV_REL] (BETA_RULE Q_IMP_equiv_Q)) 
   THEN FIRST_ASSUM MATCH_MP_TAC THEN CONV_TAC SYM_CONV
   THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL] 
  ,
   
   XRULE_ASSUM_TAC SYM THEN UNDISCH_TAC (--`R (Node (h:^plus) t)`--) 
   THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees,OUTR,Is_ltree]
   THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
   THENL
    [IMP_RES_TAC INR_ONTO2 THEN STRIP_ASSUME_TAC (ISPEC (--`(INR(INR (t':tree))):num + one + tree`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]
    ,
     FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`t':tree`--) THEN REFL_TAC
    ]
  ,
   IMP_RES_TAC IS_EL_IMP_EL THEN ASM_REWRITE_TAC[] THEN RES_TAC
   THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:((^plus)ltree)list)`--) ASSUME_TAC
   THENL[MATCH_MP_TAC EL_IS_EL THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[],ALL_TAC]
   THEN UNDISCH_TAC (--`n < LENGTH (trl:((^plus)ltree)list)`--) THEN ASM_REWRITE_TAC[]
   THEN STRIP_TAC THEN RES_TAC
   THEN IMP_RES_TAC (REWRITE_RULE[equiv_EQUIV_REL] (BETA_RULE Q_IMP_equiv_Q)) 
   THEN FIRST_ASSUM MATCH_MP_TAC THEN CONV_TAC SYM_CONV
   THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL] 
  ]);
  

  
   
val Q_pick_REP_pvt = store_thm("Q_pick_REP_pvt",
  (--`!(x:pvt). Q ((pick o REP_pvt) x)`--),
STRIP_TAC THEN STRIP_ASSUME_TAC ((BETA_RULE o (REWRITE_RULE[Is_pvt_REP,QUOTIENT_SET_def])) (ISPEC (--`x:pvt`--) EXISTS_REP_lemma))
THEN ASM_REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC THEN IMP_RES_TAC Is_pvt_REP_lemma
THEN STRIP_ASSUME_TAC (ISPEC (--`equiv (t:(^plus)ltree)`--) (CONJUNCT2 pvt_ISO_DEF))
THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN MATCH_MP_TAC (BETA_RULE Q_IMP_Q_pick_equiv) THEN ASM_REWRITE_TAC[]);



val Q_NUM_REP = store_thm("Q_NUM_REP",
   (--`!n. Q(Node ((INR (INL n)):^plus) ([]:((^plus)ltree)list))`--),
   
   REWRITE_TAC[Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,OUTR,EVERY_DEF]
   THEN REPEAT STRIP_TAC THEN IMP_RES_TAC INR_ONTO2
   THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n):num + one + tree`--) ISL_OR_ISR)
   THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL]);
   

val Q_LIST_REP = store_thm("Q_LIST_REP",
   (--`!(l:(pvt)list). Q(Node (INR (INR (INL one))) (MAP (pick o REP_pvt) l))`--),
   
   LIST_INDUCT_TAC THEN
   REWRITE_TAC[Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,OUTR,EVERY_DEF,MAP]
   THEN REPEAT STRIP_TAC
   THENL
    [IMP_RES_TAC INR_ONTO2 THEN IMP_RES_TAC (ISPEC (--`(INL one):one + tree`--) INR_ONTO2)
     THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):one + tree`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL] 
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR] 
    ,
     IMP_RES_TAC INR_ONTO2 
     THEN STRIP_ASSUME_TAC (ISPEC (--`(INR (INL one)):num + one + tree`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISL] 
     THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR] 
    ,
     IMP_RES_TAC INR_ONTO2 THEN IMP_RES_TAC (ISPEC (--`(INL one):one + tree`--) INR_ONTO2)
     THEN STRIP_ASSUME_TAC (ISPEC (--`(INL one):one + tree`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL] 
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR]
    ,
     MATCH_ACCEPT_TAC Q_pick_REP_pvt
    ,
     UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[Q_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
    ]);
    

val Q_SET_REP = store_thm("Q_SET_REP",
  (--`!(s:(pvt)set). Q(Node (INL one) (MAP (pick o REP_pvt) (S2L s)))`--),
  
   STRIP_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`(s:(pvt)set)`--) (GSYM EXISTS_LIST_FOR_S2L ))
   THEN ASM_REWRITE_TAC[] THEN SPEC_TAC ((--`tl:(pvt)list`--), (--`tl:(pvt)list`--))
   THEN POP_ASSUM (fn thm => ALL_TAC)  THEN LIST_INDUCT_TAC THEN
   REWRITE_TAC[Q_DEF,R_DEF,Tree_Root,Tree_Subtrees,OUTR,EVERY_DEF,MAP]
   THEN REPEAT STRIP_TAC
   THENL
    [STRIP_ASSUME_TAC (ISPEC (--`(INL one):^plus`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL] 
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR] 
    ,
     STRIP_ASSUME_TAC (ISPEC (--`(INL one):^plus`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL] 
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR] 
    ,
     STRIP_ASSUME_TAC (ISPEC (--`(INL one):^plus`--) ISL_OR_ISR)
     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISL] 
     THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[ISR] 
    ,
     MATCH_ACCEPT_TAC Q_pick_REP_pvt
    ,
     UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[Q_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
    ]);



val R_INR_INR_INR_REP_ltree = prove(
  (--`!(t:(pvt)ltree). R(Node (INR (INR (INR (FST (REP_ltree t))))) 
                                (MAP (pick o REP_pvt) (SND (REP_ltree t))))`--),
  
  REWRITE_TAC[R_DEF,Tree_Subtrees,Tree_Root,OUTR] THEN REPEAT STRIP_TAC
  THENL
   [IMP_RES_TAC INR_ONTO2 THEN XRULE_ASSUM_TAC SYM
    THEN STRIP_ASSUME_TAC (ISPEC (--`(INL n):num + one + tree`--) ISL_OR_ISR)
    THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[ISR] THEN ASM_REWRITE_TAC[]
    THEN REWRITE_TAC[ISL]
   ,
    MATCH_ACCEPT_TAC Is_ltree_REP_ltree_lemma
   ]);
   
    
val Q_TREE_REP_lemma = prove(
  (--`!(t:(pvt)ltree). (\t. Q(Node (INR (INR (INR (FST (REP_ltree t))))) 
                                 (MAP (pick o REP_pvt) (SND (REP_ltree t))))) t`--),
 MATCH_MP_TAC ltree_Induct THEN BETA_TAC
 THEN REWRITE_TAC[EVERY_IS_EL] THEN BETA_TAC 
 THEN REWRITE_TAC[Q_DEF,EVERY_IS_EL] 
 THEN REWRITE_TAC[R_INR_INR_INR_REP_ltree,GSYM IS_EL_MAP_EXISTS,
                  REP_Node_lemma,pairTheory.SND,IS_EL,IS_EL_FLAT]
 THEN REPEAT STRIP_TAC
 THENL
  [XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[Q_pick_REP_pvt]
  ,
   RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN UNDISCH_TAC (--`IS_EL (x':pvt) ys`--)
   THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
   THEN REWRITE_TAC[]
  ]);
  
val Q_TREE_REP = store_thm("Q_TREE_REP",
  (--`!(t:(pvt)ltree).  Q(Node (INR (INR (INR (FST (REP_ltree t))))) 
                                 (MAP (pick o REP_pvt) (SND (REP_ltree t))))`--),
  MATCH_ACCEPT_TAC (BETA_RULE Q_TREE_REP_lemma));
  


val SET_L2S_EQ_ABS = store_thm("SET_L2S_EQ_ABS",

(--`!(t:((^plus)ltree)list). 
  (EVERY Q t) ==> 
   ((SET (L2S (MAP (ABS_pvt o equiv) t))) =
      (ABS_pvt (equiv (Node (INL one) (t)))))`--),

REWRITE_TAC[SET_DEF,EVERY_IS_EL] THEN REPEAT STRIP_TAC
THEN AP_TERM_TAC THEN REWRITE_TAC[GSYM equiv_EQUIV_REL]
THEN REWRITE_TAC[equiv_CHAR] THEN DISJ2_TAC 
THEN DISJ1_TAC (* NOTE CHANGES here *)
THEN REWRITE_TAC[GSYM IS_EL_MAP_EXISTS] THEN BETA_TAC
THEN STRIP_ASSUME_TAC (ISPEC (--`(MAP (ABS_pvt o equiv) (t:((^plus)ltree)list))`--) L2S_FINITE)
 THEN IMP_RES_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV) IN_IS_EL_S2L)
 THEN POP_ASSUM (fn thm => ALL_TAC)
 THEN POP_ASSUM (fn thm => ALL_TAC) THEN ASM_REWRITE_TAC[IN_L2S_IS_EL]
 THEN POP_ASSUM (fn thm => ALL_TAC) THEN REWRITE_TAC[GSYM IS_EL_MAP_EXISTS]
 THEN REPEAT STRIP_TAC  THEN REWRITE_TAC[ISL] 
THENL [ RES_TAC THEN EXISTS_TAC (--`pick (equiv (t':(^plus)ltree))`--) 
	      THEN REWRITE_TAC[equiv_EQUIV_REL] 
	      THEN CONJ_TAC THENL[ALL_TAC , MATCH_ACCEPT_TAC (GSYM equiv_pick_equiv)]
	      THEN EXISTS_TAC (--`(ABS_pvt o equiv) (t':(^plus)ltree)`--) 
	      THEN CONJ_TAC 
	      THENL[EXISTS_TAC (--`t':(^plus)ltree`--) THEN ASM_REWRITE_TAC[],ALL_TAC]
	      THEN REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
	      THEN IMP_RES_TAC Is_pvt_REP_lemma
	      THEN STRIP_ASSUME_TAC  (ISPEC (--`equiv (t':(^plus)ltree)`--) (CONJUNCT2 pvt_ISO_DEF))
	      THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[] 
,	   
  EXISTS_TAC (--`x':((^plus)ltree)`--)
	     THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM
	     THEN ASM_REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
	     THEN RES_TAC THEN IMP_RES_TAC Is_pvt_REP_lemma
	     THEN STRIP_ASSUME_TAC (ISPEC (--`equiv (x':(^plus)ltree)`--) (CONJUNCT2 pvt_ISO_DEF))
	     THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[] 
	     THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV)
	     THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[equiv_EQUIV_REL]
	     THEN MATCH_ACCEPT_TAC (GSYM equiv_pick_equiv)
]); 
	     

val LIST_EQ_ABS = store_thm("LIST_EQ_ABS",
  (--`!(t:((^plus)ltree)list). 
  (EVERY Q t) ==> 
      (LIST (MAP (ABS_pvt o equiv) t) =
         ABS_pvt (equiv (Node (INR (INR (INL one))) t)))`--),

REWRITE_TAC[LIST_DEF,EVERY_IS_EL] THEN REPEAT STRIP_TAC
THEN AP_TERM_TAC THEN REWRITE_TAC[MAP_MAP_o]
THEN REWRITE_TAC[GSYM equiv_EQUIV_REL] THEN REWRITE_TAC[equiv_CHAR]
THEN REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
THEN DISJ2_TAC THEN DISJ2_TAC THEN REWRITE_TAC[LENGTH_MAP]
THEN REPEAT STRIP_TAC 
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`== , residue = ==`:(one + num + one + tree)ltree`==}] EL_MAP)
THEN ASM_REWRITE_TAC[] THEN BETA_TAC
THEN SUBGOAL_THEN (--`Q (EL n (t:((^plus)ltree)list))`--) ASSUME_TAC
THENL[FIRST_ASSUM MATCH_MP_TAC THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[], ALL_TAC]
THEN IMP_RES_TAC Is_pvt_REP_lemma
THEN STRIP_ASSUME_TAC (ISPEC (--`equiv ((EL n t):(^plus)ltree)`--) (CONJUNCT2 pvt_ISO_DEF))
THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN ASM_REWRITE_TAC[equiv_EQUIV_REL]
THEN MATCH_ACCEPT_TAC (GSYM equiv_pick_equiv));



val TREE_EQ_ABS = store_thm("TREE_EQ_ABS",
 
(--`!(trl:((^plus)ltree)list) (y'':tree).
  ((R (Node (INR (INR (INR y''))) trl)) /\ (EVERY Q trl)) ==>
    (TREE (ABS_ltree (y'',MAP (ABS_pvt o equiv) trl)) = 
      ABS_pvt (equiv (Node (INR (INR (INR y''))) trl)))`--),
      
REWRITE_TAC[TREE_DEF,EVERY_IS_EL]
THEN REPEAT STRIP_TAC
THEN SUBGOAL_THEN (--`Is_ltree(y'':tree, MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--) ASSUME_TAC
THENL[UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC 
      THEN REWRITE_TAC [R_DEF, Is_ltree,Tree_Root,Tree_Subtrees,LENGTH_MAP,OUTR]
      THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC
      THEN EXISTS_TAC (--`y'':tree`--) THEN REWRITE_TAC[],ALL_TAC]
THEN UNDISCH_ONE_TAC 
THEN STRIP_ASSUME_TAC 
     (ISPEC (--`((y'':tree),MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--)
      (CONJUNCT2 (ltree_ISO_DEF)))
THEN ASM_REWRITE_TAC[pairTheory.FST, pairTheory.SND] 
THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN AP_TERM_TAC
THEN REWRITE_TAC[GSYM equiv_EQUIV_REL,pairTheory.FST, pairTheory.SND] 
THEN REWRITE_TAC[equiv_CHAR]
THEN DISJ2_TAC THEN DISJ2_TAC  THEN REWRITE_TAC[LENGTH_MAP]
THEN CONJ_TAC
THENL
[
 REPEAT STRIP_TAC THEN REWRITE_TAC[MAP_MAP_o]
 THEN REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC 
 THEN IMP_RES_TAC 
      (INST_TYPE [{redex = ==`:'b`==, residue = ==`:(one + num + one + tree)ltree`==}] EL_MAP)
 THEN ASM_REWRITE_TAC[] THEN BETA_TAC 
 THEN SUBGOAL_THEN (--`Q (EL n (trl:((^plus)ltree)list))`--) ASSUME_TAC
 THENL[FIRST_ASSUM MATCH_MP_TAC THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[],ALL_TAC]
 THEN IMP_RES_TAC Is_pvt_REP_lemma
 THEN STRIP_ASSUME_TAC (ISPEC (--`equiv ((EL n trl):(^plus)ltree)`--) (CONJUNCT2 pvt_ISO_DEF))
 THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
 THEN ASM_REWRITE_TAC[equiv_EQUIV_REL]
 THEN MATCH_ACCEPT_TAC (GSYM equiv_pick_equiv)

,
 DISJ2_TAC THEN EXISTS_TAC (--`y'':tree`--) THEN ASM_REWRITE_TAC[]
]);

     
val pvt_CASES_justification = store_thm("pvt_CASES_justification",
  (--`!(x:pvt).
       (?n. (x = (NUM n)))
       \/
       (?(s:(pvt)set).  (x = (SET s)))
       \/
       (?(l:(pvt)list). x = (LIST l))
       \/
       (?(t:(pvt)ltree). x = (TREE t))`--),
      
STRIP_TAC 
THEN STRIP_ASSUME_TAC (ISPEC (--`x:pvt`--) 
                      ((BETA_RULE o (REWRITE_RULE [Is_pvt_REP,QUOTIENT_SET_def])) EXISTS_REP_lemma))
THEN UNDISCH_ONE_TAC  
THEN STRIP_ASSUME_TAC (ISPEC (--`t:(^plus)ltree`--) Node_onto)
THEN ASM_REWRITE_TAC[Q_DEF,Tree_Root,Tree_Subtrees] THEN STRIP_TAC         
THEN STRIP_ASSUME_TAC (ISPEC (--`v:^plus`--) (sumTheory.ISL_OR_ISR))
THENL
 [(* SET *)
  DISJ2_TAC  THEN DISJ1_TAC (* Note CHANGES here *)
  THEN EXISTS_TAC (--`L2S (MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--)
  THEN IMP_RES_TAC SET_L2S_EQ_ABS THEN IMP_RES_TAC sum_ISL 
  THEN STRIP_ASSUME_TAC (ISPEC (--`x':one`--) oneTheory.one)
  THEN ASM_REWRITE_TAC[]
   ,
  IMP_RES_TAC sum_ISR 
  THEN STRIP_ASSUME_TAC (ISPEC (--`y:num + one + tree`--) (sumTheory.ISL_OR_ISR))
  THENL
  [ (* NUM *)
  DISJ1_TAC THEN ASM_REWRITE_TAC[NUM_DEF]
  THEN IMP_RES_TAC sum_ISL
  THEN EXISTS_TAC (--`x':num`--) THEN ASM_REWRITE_TAC[ ]
  THEN AP_TERM_TAC   THEN AP_TERM_TAC THEN AP_TERM_TAC 
  THEN UNDISCH_TAC (--`R (Node (v:^plus) trl)`--) 
  THEN ASM_REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees]
  THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`x':num`--) THEN REWRITE_TAC[]
  ,
   IMP_RES_TAC (ISPEC (--`y:num + one + tree`--) sum_ISR) 
   THEN STRIP_ASSUME_TAC (ISPEC (--`y':one + tree`--) (sumTheory.ISL_OR_ISR))
   THENL
   [(* LIST *)
   DISJ2_TAC THEN DISJ2_TAC THEN DISJ1_TAC
   THEN IMP_RES_TAC (ISPEC (--`y':one + tree`--) sum_ISL)
   THEN STRIP_ASSUME_TAC (ISPEC (--`x':one`--) oneTheory.one)
   THEN EXISTS_TAC (--`(MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--)
   THEN IMP_RES_TAC LIST_EQ_ABS THEN ASM_REWRITE_TAC[]
   , 
   (* TREE *)
   DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC
   THEN IMP_RES_TAC (ISPEC (--`y':one + tree`--) sum_ISR)
   THEN EXISTS_TAC (--`ABS_ltree((y'':tree), (MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list)))`--)
   THEN UNDISCH_TAC (--`R (Node (v:^plus) trl)`--)
   THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
   THEN IMP_RES_TAC TREE_EQ_ABS THEN ASM_REWRITE_TAC[]
   ]
  ]
 ]);
 

val pvt_CASES_FIN_justification = prove(
  (--`!(x:pvt).
       (?n. (x = (NUM n)))
       \/
       (?(s:(pvt)set). (FINITE s) /\  (x = (SET s)))
       \/
       (?(l:(pvt)list). x = (LIST l))
       \/
       (?(t:(pvt)ltree). x = (TREE t))`--),
      
STRIP_TAC 
THEN STRIP_ASSUME_TAC (ISPEC (--`x:pvt`--) 
                      ((BETA_RULE o (REWRITE_RULE [Is_pvt_REP,QUOTIENT_SET_def])) EXISTS_REP_lemma))
THEN UNDISCH_ONE_TAC  
THEN STRIP_ASSUME_TAC (ISPEC (--`t:(^plus)ltree`--) Node_onto)
THEN ASM_REWRITE_TAC[Q_DEF,Tree_Root,Tree_Subtrees] THEN STRIP_TAC         
THEN STRIP_ASSUME_TAC (ISPEC (--`v:^plus`--) (sumTheory.ISL_OR_ISR))
THENL
 [(* SET *)
  DISJ2_TAC  THEN DISJ1_TAC (* Note CHANGES here *)
  THEN EXISTS_TAC (--`L2S (MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--)
  THEN REWRITE_TAC[L2S_FINITE]
  THEN IMP_RES_TAC SET_L2S_EQ_ABS THEN IMP_RES_TAC sum_ISL 
  THEN STRIP_ASSUME_TAC (ISPEC (--`x':one`--) oneTheory.one)
  THEN ASM_REWRITE_TAC[]
   ,
  IMP_RES_TAC sum_ISR 
  THEN STRIP_ASSUME_TAC (ISPEC (--`y:num + one + tree`--) (sumTheory.ISL_OR_ISR))
  THENL
  [ (* NUM *)
  DISJ1_TAC THEN ASM_REWRITE_TAC[NUM_DEF]
  THEN IMP_RES_TAC sum_ISL
  THEN EXISTS_TAC (--`x':num`--) THEN ASM_REWRITE_TAC[ ]
  THEN AP_TERM_TAC   THEN AP_TERM_TAC THEN AP_TERM_TAC 
  THEN UNDISCH_TAC (--`R (Node (v:^plus) trl)`--) 
  THEN ASM_REWRITE_TAC[R_DEF,Tree_Root,Tree_Subtrees]
  THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`x':num`--) THEN REWRITE_TAC[]
  ,
   IMP_RES_TAC (ISPEC (--`y:num + one + tree`--) sum_ISR) 
   THEN STRIP_ASSUME_TAC (ISPEC (--`y':one + tree`--) (sumTheory.ISL_OR_ISR))
   THENL
   [(* LIST *)
   DISJ2_TAC THEN DISJ2_TAC THEN DISJ1_TAC
   THEN IMP_RES_TAC (ISPEC (--`y':one + tree`--) sum_ISL)
   THEN STRIP_ASSUME_TAC (ISPEC (--`x':one`--) oneTheory.one)
   THEN EXISTS_TAC (--`(MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list))`--)
   THEN IMP_RES_TAC LIST_EQ_ABS THEN ASM_REWRITE_TAC[]
   , 
   (* TREE *)
   DISJ2_TAC THEN DISJ2_TAC THEN DISJ2_TAC
   THEN IMP_RES_TAC (ISPEC (--`y':one + tree`--) sum_ISR)
   THEN EXISTS_TAC (--`ABS_ltree((y'':tree), (MAP (ABS_pvt o equiv) (trl:((^plus)ltree)list)))`--)
   THEN UNDISCH_TAC (--`R (Node (v:^plus) trl)`--)
   THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
   THEN IMP_RES_TAC TREE_EQ_ABS THEN ASM_REWRITE_TAC[]
   ]
  ]
 ]);
  
val induct_lemma2 = prove( 
(--`(!(p:pvt). P p) =
      (!rep. (Is_pvt_REP rep) ==>
           (\r. (Is_pvt_REP  r) /\ ((\x. P(ABS_pvt x)) r)) rep)`--),
	   
BETA_TAC THEN EQ_TAC
THENL
 [REWRITE_TAC[REP_ONTO] THEN REPEAT STRIP_TAC
  THENL[EXISTS_TAC (--`a:pvt`--) THEN ASM_REWRITE_TAC[],
              ASM_REWRITE_TAC[pvt_ISO_DEF]]
 ,
  REPEAT STRIP_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`p:pvt`--) ABS_ONTO)
  THEN RES_TAC THEN ASM_REWRITE_TAC[]
 ]);

val induct_lemma3 = prove( 
(--`(!(p:pvt). P p) =
      (!rep t. ((rep = equiv t) /\ (Q t)) ==>
           (P o ABS_pvt o equiv) t)`--),


REWRITE_TAC[induct_lemma2,Is_pvt_REP,combinTheory.o_DEF,QUOTIENT_SET_def] THEN BETA_TAC   
THEN EQ_TAC
THENL
 [REPEAT STRIP_TAC THEN RES_TAC THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC
  THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC 
  THEN FIRST_ASSUM MATCH_ACCEPT_TAC
 ,
  REPEAT STRIP_TAC 
  THENL[ EXISTS_TAC (--`t:(^plus)ltree`--) THEN ASM_REWRITE_TAC[],
   ASM_REWRITE_TAC[] THEN RES_TAC]
 ]);
 
val induct_lemma = store_thm("induct_lemma",
(--`(!(p:pvt). P p) =
      (!t. (\t. !rep. ((rep = equiv t) /\ (Q t)) ==>
           (P o ABS_pvt o equiv) t) t)`--),


BETA_TAC THEN REWRITE_TAC[induct_lemma2,Is_pvt_REP,combinTheory.o_DEF,QUOTIENT_SET_def]
THEN BETA_TAC   
THEN EQ_TAC
THENL
 [REPEAT STRIP_TAC THEN RES_TAC THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC
  THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC 
  THEN FIRST_ASSUM MATCH_ACCEPT_TAC
 ,
  REPEAT STRIP_TAC 
  THENL[ EXISTS_TAC (--`t:(^plus)ltree`--) THEN ASM_REWRITE_TAC[],
   ASM_REWRITE_TAC[] THEN RES_TAC]
 ]);


val induct_lemma5 = prove( 
(--`(!(p:pvt). P p) =
      (!t. (\t. ((Q // equiv) (equiv t)) ==>
           (P o ABS_pvt o equiv) t) t)`--),


BETA_TAC THEN REWRITE_TAC[induct_lemma2,Is_pvt_REP,combinTheory.o_DEF,QUOTIENT_SET_def]
THEN BETA_TAC   
THEN EQ_TAC
THENL
 [REPEAT STRIP_TAC THEN RES_TAC 
 ,
  REPEAT STRIP_TAC 
  THENL[ EXISTS_TAC (--`t:(^plus)ltree`--) THEN ASM_REWRITE_TAC[],
   ASM_REWRITE_TAC[] THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC]
 ]);


(* NB: Ik heb die FINITE hierbij gedaan omdat ik dan de uniqueness van het
          para kon bewijzen! Als ik het weglaat kan ik evengoed nog gewoon
	  pvt_INDUCT bewijzen! *)


val pvt_INDUCT_justification = store_thm("pvt_INDUCT_justification",
(--`!(P:pvt -> bool).   
      (!n. P (NUM n)) /\
      (!(s:(pvt)set). ((FINITE s) /\ (!p. (p IN s) ==> (P p))) ==> (P (SET s))) /\
      (!(l:(pvt)list). (EVERY P l) ==> (P (LIST l))) /\
      (!(t:(pvt)ltree). (EVERY_TREE P t) ==> (P (TREE t)))
      ==>
      (!p. P p)`--),
      
REWRITE_TAC[EVERY_IS_EL] THEN GEN_TAC THEN STRIP_TAC 
THEN REWRITE_TAC[ induct_lemma]
THEN REPEAT STRIP_TAC
THEN MATCH_MP_TAC ltree_Induct THEN BETA_TAC
THEN REWRITE_TAC[EVERY_IS_EL] THEN REWRITE_TAC[Q_DEF] 
THEN REPEAT STRIP_TAC
THEN STRIP_ASSUME_TAC (ISPEC (--`h:^plus`--) (sumTheory.ISL_OR_ISR))
THEN REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC 
THENL
[(* SET *)
  XRULE_ASSUM_TAC (ISPEC (--`(L2S (MAP (ABS_pvt o equiv) (t:((^plus)ltree)list)))`--))
  THEN IMP_RES_TAC sum_ISL 
  THEN STRIP_ASSUME_TAC (ISPEC (--`x:one`--) oneTheory.one)
  THEN ASM_REWRITE_TAC[]  
  THEN IMP_RES_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV)  SET_L2S_EQ_ABS)
  THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
  THEN REWRITE_TAC[IN_L2S_IS_EL, GSYM IS_EL_MAP_EXISTS,L2S_FINITE]
  THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) 
  THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
  THEN XRULE_ASSUM_TAC (BETA_RULE o REWRITE_RULE[EVERY_IS_EL,combinTheory.o_DEF])
  THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`equiv (x':(^plus)ltree)`--)
  THEN REFL_TAC
, 
  IMP_RES_TAC sum_ISR 
  THEN STRIP_ASSUME_TAC (ISPEC (--`y:num + one + tree`--) (sumTheory.ISL_OR_ISR))
  THENL
[ (* NUM *)
 IMP_RES_TAC sum_ISL 
 THEN UNDISCH_TAC (--`R (Node (h:^plus) t)`--) THEN ASM_REWRITE_TAC[]
 THEN STRIP_TAC THEN IMP_RES_TAC R_INR_INL
 THEN ASM_REWRITE_TAC[GSYM NUM_DEF] 
 
 ,
  IMP_RES_TAC (ISPEC (--`y:num + one + tree`--) sum_ISR)
  THEN STRIP_ASSUME_TAC (ISPEC (--`y':one + tree`--) (sumTheory.ISL_OR_ISR))
 THENL
  [(* LIST *)
  XRULE_ASSUM_TAC (ISPEC (--`((MAP (ABS_pvt o equiv) (t:((^plus)ltree)list)))`--))
  THEN IMP_RES_TAC sum_ISL 
  THEN STRIP_ASSUME_TAC (ISPEC (--`x:one`--) oneTheory.one)
  THEN ASM_REWRITE_TAC[EVERY_IS_EL] 
  THEN IMP_RES_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV)  LIST_EQ_ABS)
  THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
  THEN REWRITE_TAC[ GSYM IS_EL_MAP_EXISTS]
  THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) 
  THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
  THEN XRULE_ASSUM_TAC (BETA_RULE o REWRITE_RULE[EVERY_IS_EL,combinTheory.o_DEF])
  THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`equiv (x'':(^plus)ltree)`--)
  THEN REFL_TAC
  ,
  (* TREE *)
  IMP_RES_TAC (ISPEC (--`y':one + tree`--) sum_ISR)
  THEN  ASM_REWRITE_TAC[]
  THEN XRULE_ASSUM_TAC (ISPEC (--`ABS_ltree (y'':tree,MAP (ABS_pvt o equiv) (t:((^plus)ltree)list))`--))
  THEN UNDISCH_TAC (--`R (Node (h:^plus) t)`--) THEN ASM_REWRITE_TAC[] THEN STRIP_TAC
  THEN IMP_RES_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV) TREE_EQ_ABS)
  THEN ASM_REWRITE_TAC[] THEN  FIRST_ASSUM MATCH_MP_TAC 
  THEN REWRITE_TAC[EVERY_TREE] 
  THEN SUBGOAL_THEN (--`Is_ltree (y'':tree,MAP (ABS_pvt o equiv) (t:((^plus)ltree)list))`--) ASSUME_TAC
  THENL[IMP_RES_TAC R_INR_INR_INR THEN UNDISCH_ONE_TAC 
  THEN REWRITE_TAC[pairTheory.SND,Is_ltree,LENGTH_MAP],UNDISCH_ONE_TAC]
  THEN REWRITE_TAC[pairTheory.SND,CONJUNCT2 (ltree_ISO_DEF)]
  THEN STRIP_TAC THEN ASM_REWRITE_TAC[pairTheory.SND,EVERY_IS_EL,GSYM IS_EL_MAP_EXISTS]
  THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN REPEAT STRIP_TAC
  THEN XRULE_ASSUM_TAC (BETA_RULE o REWRITE_RULE[EVERY_IS_EL,combinTheory.o_DEF]) 
  THEN ASM_REWRITE_TAC[] THEN RES_TAC
  THEN FIRST_ASSUM MATCH_MP_TAC THEN EXISTS_TAC (--`equiv (x':(^plus)ltree)`--)
  THEN REFL_TAC
  ]
 ]
 
] ); 



val thm1_1 = GEN_ALL (CONJUNCT1 (SPEC_ALL 
                                       (BETA_RULE 
				         (REWRITE_RULE [EXISTS_UNIQUE_DEF] ltree_Axiom))));
val tm =
     --`\l:('b)list.
        \v:(^plus). 
        \ps:((^plus)ltree)list.
      (ISL v => fset (L2S (ZIP (l, (MAP (ABS_pvt o equiv) ps)))) 
             |  (ISL (OUTR v)) => fnum (OUTL (OUTR v)) 
	     |  (ISL (OUTR (OUTR v))) => flist (ZIP(l, (MAP (ABS_pvt o equiv) ps)))
	     | ftree (ZIP_TREE((ABS_ltree ((OUTR (OUTR (OUTR v))), l)), 
	                (ABS_ltree ((OUTR (OUTR (OUTR v))),(MAP (ABS_pvt o equiv) ps)))))
            ):'b`--; 

val thm1 = BETA_RULE (ISPEC tm thm1_1);

val thm1uniq =  BETA_RULE (ISPEC tm  (GEN_ALL (CONJUNCT2 (SPEC_ALL 
                           (BETA_RULE 
				         (REWRITE_RULE [EXISTS_UNIQUE_DEF] ltree_Axiom))))));
	 					 


val thm2_1 = (ISPEC (--`(equiv :(^plus)ltree -> (^plus)ltree -> bool)`--) QUOTIENT_THM);
val thm2_2 = MATCH_MP thm2_1 equiv_EQUIV_REL;
val thm2_3 = (ISPEC (--`REP_pvt`--) (ISPEC (--`ABS_pvt`--) 
              (ISPEC (--`Q:(^plus)ltree->bool`--) thm2_2)));
val thm2 = MATCH_MP  thm2_3 (REWRITE_RULE[Is_pvt_REP] pvt_ISO_DEF);



val TREE_LIST_thm3_TAC =
 MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP]
THEN CONJ_TAC THENL[ASM_REWRITE_TAC[],ALL_TAC] THEN REPEAT STRIP_TAC
THEN RES_TAC
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`: 'b`== ,  residue = ==`:pvt`==}] 
                                 (EL_MAP))
THEN ASM_REWRITE_TAC[]	THEN POP_ASSUM (fn thm => ALL_TAC)  
THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
THEN REPEAT STRIP_TAC 
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`: 'b`== ,  residue = ==`:pvt`==}] 
                                 (EL_MAP))
THEN ASM_REWRITE_TAC[]	
THEN REWRITE_TAC[combinTheory.o_DEF] THEN BETA_TAC
THEN XRULE_ASSUM_TAC (REWRITE_RULE [equiv_EQUIV_REL])
THEN ASM_REWRITE_TAC[];

fun DO_TAC tac n = 
    if (n = 0)
    then ALL_TAC
    else tac THEN (DO_TAC tac (n-1));




val REWRITE_ZIP_LENGTH_TAC_thm3 = 

SUBGOAL_THEN (--`(LENGTH (ZIP ((xs1:('b)list),MAP (equiv:(^plus)ltree->(^plus)ltree->bool) tl1))) 
                            = (LENGTH xs1)`--) ASSUME_TAC
THENL[MATCH_MP_TAC  LENGTH_ZIP_L THEN ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
THEN SUBGOAL_THEN (--`(LENGTH (ZIP ((xs2:('b)list),MAP (equiv:(^plus)ltree->(^plus)ltree->bool) tl2))) 
                            = (LENGTH xs2)`--) ASSUME_TAC
THENL[MATCH_MP_TAC  LENGTH_ZIP_L THEN ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
THEN ASM_REWRITE_TAC[EXTENSION,IN_L2S_IS_EL,IS_EL_EQ_EL] 
THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC)
THEN SUBGOAL_THEN (--`(LENGTH (ZIP ((xs1:('b)list),
                                  MAP (ABS_pvt o (equiv:(^plus)ltree->(^plus)ltree->bool)) tl1))) 
                            = (LENGTH xs1)`--) ASSUME_TAC
THENL[MATCH_MP_TAC  LENGTH_ZIP_L THEN ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
THEN SUBGOAL_THEN (--`(LENGTH (ZIP ((xs2:('b)list),
                 MAP (ABS_pvt o (equiv:(^plus)ltree->(^plus)ltree->bool)) tl2))) 
                            = (LENGTH xs2)`--) ASSUME_TAC
THENL[MATCH_MP_TAC  LENGTH_ZIP_L THEN ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
THEN SUBGOAL_THEN (--`!x. (?n. (n < (LENGTH (tl1:((^plus)ltree)list)))
                         /\ ((x:('b # ((^plus)ltree -> bool))) = EL n (ZIP (xs1,MAP equiv tl1))))
                       =
                       (?n. (n < (LENGTH tl1)) /\ (x = ((EL n xs1), (EL n (MAP equiv tl1)))))`--)
                    ASSUME_TAC
THENL
 [GEN_TAC THEN EQ_TAC THEN STRIP_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
  THENL[ALL_TAC, CONV_TAC SYM_CONV] THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
 ,ALL_TAC]
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
THEN SUBGOAL_THEN (--`!x. (?n. (n < (LENGTH (tl2:((^plus)ltree)list)))
                         /\ ((x:('b # ((^plus)ltree -> bool))) = EL n (ZIP (xs2,MAP equiv tl2))))
                       =
                       (?n. (n < (LENGTH tl2)) /\ (x = ((EL n xs2), (EL n (MAP equiv tl2)))))`--)
                    ASSUME_TAC
THENL
 [GEN_TAC THEN EQ_TAC THEN STRIP_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
  THENL[ALL_TAC, CONV_TAC SYM_CONV] THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
 ,ALL_TAC]
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
THEN ASM_REWRITE_TAC[EQ_IMP_THM] THEN CONV_TAC (ONCE_DEPTH_CONV FORALL_AND_CONV)
THEN STRIP_TAC THEN REWRITE_TAC[GSYM MAP_MAP_o]
THEN SUBGOAL_THEN (--`!x. (?n. (n < (LENGTH (tl1:((^plus)ltree)list)))
                         /\ ((x:('b # pvt)) = EL n (ZIP (xs1,(MAP ABS_pvt (MAP equiv tl1))))))
                       =
                       (?n. (n < (LENGTH tl1)) /\ (x = ((EL n xs1), (ABS_pvt (EL n (MAP equiv tl1))))))`--)
                    ASSUME_TAC
THENL
 [GEN_TAC 
  THEN EQ_TAC THEN STRIP_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
  THEN 
    let 
     val tac = SUBGOAL_THEN (--`n < (LENGTH (MAP (equiv:(^plus)ltree->(^plus)ltree->bool) tl1))`--) ASSUME_TAC
              THENL[ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
    in tac 
    end
  THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:pvt`==}] (GSYM EL_MAP))
  THEN ASM_REWRITE_TAC[] THENL[ALL_TAC , CONV_TAC SYM_CONV]
  THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
 ,ALL_TAC] 
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
THEN SUBGOAL_THEN (--`!x. (?n. (n < (LENGTH (tl2:((^plus)ltree)list)))
                         /\ ((x:('b # pvt)) = EL n (ZIP (xs2,(MAP ABS_pvt (MAP equiv tl2))))))
                       =
                       (?n. (n < (LENGTH tl2)) /\ (x = ((EL n xs2), (ABS_pvt (EL n (MAP equiv tl2))))))`--)
                    ASSUME_TAC
THENL
 [GEN_TAC 
  THEN EQ_TAC THEN STRIP_TAC THEN EXISTS_TAC (--`n:num`--) THEN ASM_REWRITE_TAC[]
  THEN 
    let 
     val tac = SUBGOAL_THEN (--`n < (LENGTH (MAP (equiv:(^plus)ltree->(^plus)ltree->bool) tl2))`--) ASSUME_TAC
              THENL[ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
    in tac 
    end
  THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:pvt`==}] (GSYM EL_MAP))
  THEN ASM_REWRITE_TAC[] THENL[ALL_TAC , CONV_TAC SYM_CONV]
  THEN MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
 ,ALL_TAC] 
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC);




val thm3 = prove(
(--`(!(xs1:('b)list) (xs2:('b)list) (tl1:((^plus)ltree)list) (tl2:((^plus)ltree)list) (v:(^plus)).
         equiv (Node v tl1) (Node v tl2) /\
	 ((ISL v) ==> (((LENGTH xs1) = (LENGTH tl1)) /\ ((LENGTH xs2) = (LENGTH tl2)) /\ 
	       ((L2S (ZIP(xs1,(MAP equiv tl1)))) = (L2S (ZIP(xs2,(MAP equiv tl2))))))) /\
	 ((ISR v) ==> (xs1 = xs2)) ==>
         (^tm xs1 v tl1 = ^tm xs2 v tl2))`--),
BETA_TAC THEN REWRITE_TAC[equiv_CHAR]
THEN REPEAT STRIP_TAC THEN RES_TAC
THEN ASM_REWRITE_TAC[ISL,ISR,OUTL,OUTR]
THENL
[ (* SET *)
AP_TERM_TAC THEN UNDISCH_ONE_TAC 
THEN REWRITE_ZIP_LENGTH_TAC_thm3
THEN REPEAT STRIP_TAC 
THENL
 [UNDISCH_ONE_TAC THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # pvt)`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN SUBGOAL_THEN (--`?y. y = (EL n (MAP (equiv:(^plus)ltree -> (^plus)ltree -> bool) tl1))`--)
      STRIP_ASSUME_TAC
  THENL[EXISTS_TAC (--`(EL n (MAP (equiv:(^plus)ltree -> (^plus)ltree -> bool) tl1))`--) THEN REFL_TAC, ALL_TAC]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN XRULE_ASSUM_TAC SYM THEN STRIP_TAC
  THEN XRULE_ASSUM_TAC ((REWRITE_RULE[pairTheory.PAIR_EQ]) o 
       (ISPEC (--`((FST (x:('b # pvt))),(y:(^plus)ltree -> bool))`--)))
  THEN RES_TAC THEN EXISTS_TAC (--`n'''':num`--)
  THEN ASM_REWRITE_TAC[] 
 ,
  UNDISCH_ONE_TAC THEN STRIP_ASSUME_TAC (GSYM (ISPEC (--`x:('b # pvt)`--) (pairTheory.PAIR)))
  THEN ONCE_ASM_REWRITE_TAC[] THEN REWRITE_TAC[pairTheory.PAIR_EQ]
  THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN SUBGOAL_THEN (--`?y. y = (EL n (MAP (equiv:(^plus)ltree -> (^plus)ltree -> bool) tl2))`--)
      STRIP_ASSUME_TAC
  THENL[EXISTS_TAC (--`(EL n (MAP (equiv:(^plus)ltree -> (^plus)ltree -> bool) tl2))`--) THEN REFL_TAC, ALL_TAC]
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN XRULE_ASSUM_TAC SYM THEN STRIP_TAC
  THEN XRULE_ASSUM_TAC ((REWRITE_RULE[pairTheory.PAIR_EQ]) o 
       (ISPEC (--`((FST (x:('b # pvt))),(y:(^plus)ltree -> bool))`--)))
  THEN RES_TAC THEN EXISTS_TAC (--`n'''':num`--)
  THEN ASM_REWRITE_TAC[] 

 ]
,
(* LIST *)
AP_TERM_TAC THEN MATCH_MP_TAC ZIP_EQ
THEN CONJ_TAC 
THENL[FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[ISR], TREE_LIST_thm3_TAC]
,
 (* TREE *)

AP_TERM_TAC
THEN MATCH_MP_TAC ZIP_TREE_EQ
THEN CONJ_TAC THEN  AP_TERM_TAC THEN REWRITE_TAC[pairTheory.PAIR_EQ] 
THENL[FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[ISR], TREE_LIST_thm3_TAC]
] 
);



		 
val thm4_1 = (BETA_RULE ltree_Axiom_PRESERVES_equiv);
val thm4_2 = BETA_RULE (ISPEC tm (ISPEC (--`fn':(^plus) ltree -> 'b`--) 
                     (ISPEC (--`x:(^plus) ltree`--) (ISPEC (--`y:(^plus) ltree`--) thm4_1))));
val thm4 = GEN_ALL thm4_2;


val thm5 = prove(
   (--`!(ftree:(('b # pvt))ltree ->'b)
       (flist:(('b # pvt))list ->'b) (fset:(('b # pvt))set ->'b) 
          (fnum:num->'b) (fn':(^plus)ltree->'b).
	   (!v tl.
            fn' (Node v tl) =
              ((ISL v)
               => (fset (L2S (ZIP((MAP fn' tl),(MAP (ABS_pvt o equiv) tl)))))
                | (ISL (OUTR v)) => (fnum (OUTL (OUTR v)))
		| (ISL (OUTR (OUTR v))) => (flist (ZIP((MAP fn' tl), (MAP (ABS_pvt o equiv) tl))))
		| (ftree (ZIP_TREE((ABS_ltree (OUTR (OUTR (OUTR v)), (MAP fn' tl))),
		            (ABS_ltree (OUTR (OUTR (OUTR v)),(MAP (ABS_pvt o equiv) tl))))))))
      ==>
      (!(x:(^plus)ltree) y . (equiv x y) ==> (fn' x = fn' y))`--),
	
   REPEAT STRIP_TAC THEN CONV_TAC SYM_CONV THEN MATCH_MP_TAC thm4
   THEN EXISTS_TAC (--`ftree:(('b # pvt))ltree ->'b`--)
   THEN EXISTS_TAC (--`flist:(('b # pvt))list ->'b`--)
   THEN EXISTS_TAC (--`(fnum:num->'b)`--) 
   THEN EXISTS_TAC (--`fset:(('b # pvt))set ->'b`--)
   THEN ASM_REWRITE_TAC[(BETA_RULE thm3)]
   THEN REWRITE_TAC[equiv_EQUIV_REL] THEN CONV_TAC SYM_CONV
   THEN ASM_REWRITE_TAC[GSYM equiv_EQUIV_REL] );
   


val pvt_Axiom_justification = store_thm("pvt_Axiom_justification",

(--`!(fnum:num->'b) (fset:(('b # pvt))set ->'b) 
     (flist:(('b # pvt)list) -> 'b) (ftree:(('b # pvt)ltree) -> 'b).
     ?! (para:pvt->'b).
         (!n. (para (NUM n) = (fnum n)))
       /\
         (!s. (FINITE s) ==> (para (SET s) = (fset (IMAGE (splid para) s))))
       /\
         (!l. (para (LIST l) = flist (MAP (splid para) l)))
       /\
         (!t. (para (TREE t) = ftree (MAP_TREE (splid para) t)))`--),

  REPEAT GEN_TAC THEN STRIP_ASSUME_TAC thm1
  THEN IMP_RES_TAC thm5 
  THEN XRULE_ASSUM_TAC (ONCE_REWRITE_RULE [equiv_EQUIV_REL])
  THEN XRULE_ASSUM_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV))
  THEN XRULE_ASSUM_TAC (ONCE_REWRITE_RULE [GSYM equiv_EQUIV_REL])
  THEN IMP_RES_TAC thm2 THEN UNDISCH_ONE_TAC 
  THEN XRULE_ASSUM_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV))
  THEN REWRITE_TAC[EXISTS_UNIQUE_DEF] THEN BETA_TAC THEN BETA_TAC
  THEN STRIP_TAC THEN CONJ_TAC
  THENL
   [EXISTS_TAC (--`g:pvt->'b`--) THEN POP_ASSUM (fn thm => ALL_TAC)
    THEN ASM_REWRITE_TAC[NUM_DEF,SET_DEF,LIST_DEF,TREE_DEF]
    THEN REPEAT STRIP_TAC
    THENL
     [(* NUM *)
      STRIP_ASSUME_TAC (ISPEC (--`n:num`--) Q_NUM_REP) THEN RES_TAC
      THEN ASM_REWRITE_TAC[ISL,OUTL,OUTR,ISR]
     ,
      (* SET *)
      STRIP_ASSUME_TAC (ISPEC (--`(s:(pvt)set)`--) Q_SET_REP) THEN RES_TAC
      THEN ASM_REWRITE_TAC[ISL,OUTL,OUTR,ISR]
      THEN REWRITE_TAC[MAP_MAP_o,combinTheory.o_DEF] THEN BETA_TAC 
      THEN REWRITE_TAC[equiv_pick_REP_pvt, pvt_ISO_DEF] 
      THEN REWRITE_TAC[MAP_ID] THEN REWRITE_TAC[ZIP_MAP_EQ_MAP_splid]
      THEN REWRITE_TAC[L2S_MAP_eq_IMAGE] THEN IMP_RES_TAC L2S_S2L_id THEN ASM_REWRITE_TAC[]
      THEN AP_TERM_TAC
      THEN REWRITE_TAC[EXTENSION,IN_IMAGE,splid_def] THEN BETA_TAC
      THEN STRIP_TAC 
      THEN SUBGOAL_THEN (--`!x. (x IN (s:(pvt)set)) ==> (((g:pvt->'b) x) = (fn (pick (REP_pvt x))))`--) ASSUME_TAC
      THENL
         [UNDISCH_TAC (--`Q (Node (INL one) (MAP (pick o REP_pvt) (S2L (s:(pvt)set))))`--)
          THEN REWRITE_TAC[Q_DEF,Tree_Root, Tree_Subtrees,EVERY_IS_EL,GSYM IS_EL_MAP_EXISTS]
          THEN IMP_RES_TAC (CONV_RULE (ONCE_DEPTH_CONV SYM_CONV) IN_IS_EL_S2L) THEN ASM_REWRITE_TAC[]
          THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC) THEN POP_ASSUM (fn thm => ALL_TAC)
          THEN REPEAT STRIP_TAC
          THEN XRULE_ASSUM_TAC (ISPEC (--`pick (REP_pvt x)`--)) THEN RES_TAC 
          THEN XRULE_ASSUM_TAC 
               ((REWRITE_RULE []) o BETA_RULE o (REWRITE_RULE[combinTheory.o_DEF,equiv_pick_REP_pvt, pvt_ISO_DEF]))
          THEN RES_TAC
         ,
          EQ_TAC THEN REPEAT STRIP_TAC THEN RES_TAC
          THEN EXISTS_TAC (--`x':pvt`--) THEN ASM_REWRITE_TAC[]
         ]
     ,
      (* LIST *)
      STRIP_ASSUME_TAC (ISPEC (--`(l:(pvt)list)`--) Q_LIST_REP) THEN RES_TAC
      THEN ASM_REWRITE_TAC[ISL,OUTL,OUTR,ISR]
      THEN REWRITE_TAC[MAP_MAP_o,combinTheory.o_DEF] THEN BETA_TAC
      THEN REWRITE_TAC[equiv_pick_REP_pvt, pvt_ISO_DEF,ETA_AX,MAP_ID] THEN REWRITE_TAC[ZIP_MAP_EQ_MAP_splid]
      THEN AP_TERM_TAC
      THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP]
      THEN REPEAT STRIP_TAC 
      THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:('b # pvt)`==}] EL_MAP) 
      THEN ASM_REWRITE_TAC[splid_def] THEN BETA_TAC THEN REWRITE_TAC[pairTheory.PAIR_EQ]
      THEN SUBGOAL_THEN (--`!x. (IS_EL x (l:(pvt)list)) ==> (((g:pvt->'b) x) = (fn (pick (REP_pvt x))))`--) ASSUME_TAC
        THENL
         [UNDISCH_TAC (--`Q (Node (INR (INR (INL one))) (MAP (pick o REP_pvt) (l:(pvt)list)))`--)
          THEN REWRITE_TAC[Q_DEF,Tree_Root, Tree_Subtrees,EVERY_IS_EL,GSYM IS_EL_MAP_EXISTS]
          THEN REPEAT STRIP_TAC
          THEN XRULE_ASSUM_TAC (ISPEC (--`pick (REP_pvt x)`--)) THEN RES_TAC 
          THEN XRULE_ASSUM_TAC 
               ((REWRITE_RULE []) o BETA_RULE o (REWRITE_RULE[combinTheory.o_DEF,equiv_pick_REP_pvt, pvt_ISO_DEF]))
          THEN RES_TAC
         ,
          CONV_TAC SYM_CONV THEN FIRST_ASSUM MATCH_MP_TAC
          THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[]
         ]
     ,
      (* TREE *)
      STRIP_ASSUME_TAC (ISPEC (--`(t:(pvt)ltree)`--) Q_TREE_REP) THEN RES_TAC
      THEN ASM_REWRITE_TAC[ISL,OUTL,OUTR,ISR]
      THEN REWRITE_TAC[MAP_MAP_o,combinTheory.o_DEF] THEN BETA_TAC
      THEN REWRITE_TAC[equiv_pick_REP_pvt, pvt_ISO_DEF,ETA_AX,MAP_ID]
      THEN RW_TAC std_ss []
      THEN REWRITE_TAC[GSYM MAP_TREE_ABS,ltree_ISO_DEF,pairTheory.FST,pairTheory.SND]
      THEN REWRITE_TAC[ZIP_TREE_MAP_TREE_EQ_MAP_TREE_splid]
      THEN REWRITE_TAC[MAP_TREE_ABS] THEN AP_TERM_TAC THEN AP_TERM_TAC
      THEN REWRITE_TAC[pairTheory.PAIR_EQ]
      THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP,pairTheory.FST,pairTheory.SND]
      THEN REPEAT STRIP_TAC 
      THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:('b # pvt)`==}] EL_MAP)
      THEN ASM_REWRITE_TAC[splid_def] THEN BETA_TAC
      THEN REWRITE_TAC[pairTheory.PAIR_EQ]
      THEN SUBGOAL_THEN (--`!x. (IS_EL x (SND (REP_ltree (t:(pvt)ltree))))
                             ==> (((g:pvt->'b) x) = (fn (pick (REP_pvt x))))`--) ASSUME_TAC
      THENL
         [UNDISCH_TAC (--`Q (Node (INR (INR (INR (FST (REP_ltree (t:(pvt)ltree)))))) 
                            (MAP (pick o REP_pvt) (SND (REP_ltree (t:(pvt)ltree)))))`--)
          THEN REWRITE_TAC[Q_DEF,Tree_Root, Tree_Subtrees,EVERY_IS_EL,GSYM IS_EL_MAP_EXISTS]
          THEN REPEAT STRIP_TAC
          THEN XRULE_ASSUM_TAC (ISPEC (--`pick (REP_pvt x)`--)) THEN RES_TAC 
          THEN XRULE_ASSUM_TAC 
               ((REWRITE_RULE []) o BETA_RULE o (REWRITE_RULE[combinTheory.o_DEF,equiv_pick_REP_pvt, pvt_ISO_DEF]))
          THEN RES_TAC
         ,
          CONV_TAC SYM_CONV THEN FIRST_ASSUM MATCH_MP_TAC
          THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[]
         ]
     ]
 
   ,
    (* UNIQUENESS PART *)
    REPEAT STRIP_TAC
    THEN CONV_TAC FUN_EQ_CONV 
    THEN SUBGOAL_THEN (--`!(p:pvt). (\p. (x:pvt->'b) p = y p) p`--) ASSUME_TAC
    THENL
     [MATCH_MP_TAC pvt_INDUCT_justification THEN BETA_TAC THEN REPEAT STRIP_TAC
      THEN RES_TAC THEN ASM_REWRITE_TAC[]
      THENL
      [(* SET *)
       AP_TERM_TAC
       THEN REWRITE_TAC[EXTENSION, IN_IMAGE,splid_def]
       THEN STRIP_TAC THEN EQ_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[]
       THEN EXISTS_TAC (--`x'' : pvt`--) THEN ASM_REWRITE_TAC[] THEN RES_TAC
       THEN ASM_REWRITE_TAC[]
       
       ,
       (* LIST *)
       AP_TERM_TAC
       THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP]
       THEN REPEAT STRIP_TAC 
       THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:('b # pvt)`==}]  (EL_MAP))
       THEN ASM_REWRITE_TAC[splid_def,pairTheory.PAIR_EQ] 
       THEN XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE[EVERY_IS_EL]))
       THEN FIRST_ASSUM MATCH_MP_TAC
       THEN MATCH_MP_TAC EL_IS_EL
       THEN FIRST_ASSUM MATCH_ACCEPT_TAC
       
       ,
        (* TREE *)
      AP_TERM_TAC THEN MATCH_MP_TAC EVERY_TREE_IMP_MAP_TREE_EQ
      THEN ASM_REWRITE_TAC[splid_def,pairTheory.PAIR_EQ] 
      ]
     ,
      XRULE_ASSUM_TAC BETA_RULE THEN ASM_REWRITE_TAC[]
    ]
  ]);
  
         

val _ = export_theory ();





