
open HolKernel Parse basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ; 

val _ = new_theory "quotient";


(*-------------------------------------------------*)
(* Tom's quotients             			   *)
(*-------------------------------------------------*)

(*

load "bossLib" ;

*)

open bossLib;


val EQUIV_DEF = store_thm("EQUIV_DEF",
     (--`!(R:'a->'a->bool).
            (!x y. (R x y) = (R x = R y))
	    =
	    ((!x. R x x)
	    /\
	     (!x y. R x y ==> R y x)
	    /\
	     (!x y z. (R x y /\ R y z) ==> R x z))`--),
REPEAT (STRIP_TAC ORELSE EQ_TAC)
THEN RES_TAC THEN ASM_REWRITE_TAC[]
THEN CONV_TAC FUN_EQ_CONV THEN GEN_TAC
THEN EQ_TAC THEN STRIP_TAC THEN RES_TAC);


val QUOTIENT_SET = xDefine "QUOTIENT_SET" `// (Q:'a->bool) (E:'a->'a->bool) = (\s. ?t. (s = (E t)) /\ (Q t))`;
  
val _ = set_fixity "//" (Infixl 600); 

val LEMMA1 =
    prove
      (--`!(g:'a->'b). !(f:'a->'c).
              (!x y. (g x = g y) ==> (f x = f y)) ==>
	       !(b:'b). ?(c:'c). !(a:'a). 
	            (b = g a) ==> (c = f a)`--,
    REPEAT STRIP_TAC
    THEN let 
                val cases = SPEC (--`?(x:'a). (b:'b) = g x`--)  EXCLUDED_MIDDLE
		val cthm = CONV_RULE (ONCE_DEPTH_CONV NOT_EXISTS_CONV) cases
             in
	       STRIP_ASSUME_TAC cthm
	     end 
    THENL
     [EXISTS_TAC (--`(f (x:'a)):'c`--) THEN ASM_REWRITE_TAC[]
     ,ASM_REWRITE_TAC[]]);
     
val LEMMA2 = CONV_RULE (ONCE_DEPTH_CONV (X_SKOLEM_CONV (--`h:'b->'c`--))) LEMMA1;
			 	       
		    
val LEMMA = 
     prove
      (--`!(g:'a->'b). !(f:'a->'c).
            (!x y. (g x = g y) ==> (f x = f y))
	    ==>
	    (?(h:'b->'c). !(x:'a). h(g x) = (f x))`--,
	    
    REPEAT STRIP_TAC THEN   IMP_RES_TAC LEMMA2
    THEN EXISTS_TAC (--`h:'b->'c`--)
    THEN GEN_TAC THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REFL_TAC);
     

val lemma1 =
   prove
    (--`!(R:'a->'a->bool). ?x a. x = (R a)`--,
    GEN_TAC THEN EXISTS_TAC (--`(R:'a->'a->bool) a`--)
    THEN EXISTS_TAC (--`a:'a`--) THEN REFL_TAC);
    
val lemma2 =
   prove
     (--`(!y:'a->bool. (?a:'a. y = R a) = (rep (abs y:'b) = y)) ==>
            !x:'a. rep(abs (R x)) = R x`--,
	    
     REPEAT STRIP_TAC THEN POP_ASSUM (fn th => REWRITE_TAC[GSYM th])
     THEN EXISTS_TAC (--`x:'a`--) THEN REFL_TAC);
     
val lemma3  =
   prove
    (--`(!x:'b. abs(rep x:'a->bool) = x) ==>
           (!y:'a->bool. (?a:'a. y = R a) = (rep (abs y:'b) = y)) ==>
	   !b:'b. ?x:'a. rep b = R x`--,
	   
     REPEAT STRIP_TAC THEN POP_ASSUM (MP_TAC o SPEC (--`(rep:'b->'a->bool) b`--))
     THEN ASM_REWRITE_TAC[]);


val XRULE_ASSUM_TAC = 
       let 
        fun app (rule:thm->thm) = (fn (t:thm) => (rule t) handle _ =>  t)
       in
       RULE_ASSUM_TAC o app
       end;

val LEMMA3  =
   prove
    (--`!(Q:'a -> bool) (E:'a ->'a->bool).
      !(abs:('a->bool)->'b) (rep:'b->('a->bool)).
        ((!a. (abs (rep a)) = a)
	/\
	(!r. ((Q // E) r) = (rep (abs r) = r)))
	==>
      (!y. ?x. ((rep y) = (E x)) /\ (Q x))`--,
	   
     REWRITE_TAC[QUOTIENT_SET] THEN BETA_TAC THEN REPEAT STRIP_TAC  
     THEN XRULE_ASSUM_TAC (ISPEC (--`(rep:'b->('a->bool)) y`--))
     THEN ASM_REWRITE_TAC[]);

val UNDISCH_ONE_TAC =
    let
        fun th_tac (th:thm) (tac:tactic) = (UNDISCH_TAC (concl th)) THEN tac
        fun u_asm  (th:thm) = itlist  th_tac [th] ALL_TAC
    in
        FIRST_ASSUM u_asm
    end;
  	   
val QUOTIENT_THM = 
    store_thm("QUOTIENT_THM",
      (--`!(E:'a->'a->bool).
             (!x y. E x y = (E x = E y))
	        ==>
	     (!(Q:'a->bool) (abs:('a->bool)->'b) (rep:'b->('a->bool)).
	      ((!a. (abs (rep a)) = a)
	       /\
	       (!r. ((Q // E) r) = (rep (abs r) = r)))
	      ==>
       !(f:'a->'c).
	  (!x y. (E x y) ==> (f x = f y))
	  ==>
	  (?! (g:'b -> 'c).
	   !x. (Q x) ==> (g (abs (E x)) = f x)))`--),
   
   GEN_TAC THEN REWRITE_TAC[QUOTIENT_SET]
   THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN BETA_TAC
   THEN REPEAT STRIP_TAC  
   THEN CONV_TAC EXISTS_UNIQUE_CONV THEN REPEAT STRIP_TAC
     THENL
       [IMP_RES_TAC LEMMA
        THEN EXISTS_TAC (--`\x:'b. (h ((rep x):'a->bool)):'c`--)
	THEN BETA_TAC THEN REPEAT STRIP_TAC THEN RES_TAC
	THEN SUBGOAL_THEN (--`rep ((abs:('a->bool)->'b) (E (x:'a))) = ((E:'a->'a->bool) x)`--) ASSUME_TAC
	THENL
	 [XRULE_ASSUM_TAC (SYM o (ISPEC (--`((E:'a->'a->bool) x)`--))) THEN ASM_REWRITE_TAC[]
	  THEN EXISTS_TAC (--`x:'a`--) THEN ASM_REWRITE_TAC[]
	 ,
	  ASM_REWRITE_TAC[]
	 ]
       ,
        CONV_TAC FUN_EQ_CONV THEN GEN_TAC
	 THEN 
	   let
	    val thm1 = ASSUME (--`!x:'b. abs(rep x:'a->bool) = x`--)
	   in
	    SUBST1_TAC (SYM (SPEC (--`x:'b`--) thm1)) 
	   end
	  THEN IMP_RES_TAC ((BETA_RULE o (REWRITE_RULE [QUOTIENT_SET])) LEMMA3) 
	  THEN XRULE_ASSUM_TAC (SPEC (--`x:'b`--))
	  THEN UNDISCH_ONE_TAC THEN STRIP_TAC THEN RES_TAC
	  THEN UNDISCH_TAC (--`abs (rep (x:'b):'a->bool) = x`--) 
	  THEN ASM_REWRITE_TAC []
       ]);
      

val _ = export_theory();
